# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import eventlet
import mock

from oslo_config import cfg
from oslo_utils import timeutils
from oslo_utils import uuidutils

from senlin.common import consts
from senlin.common import context
from senlin.common import messaging as rpc_messaging
from senlin.engine import service
from senlin.objects import service as service_obj
from senlin.tests.unit.common import base


@mock.patch('senlin.engine.dispatcher.Dispatcher')
@mock.patch('senlin.engine.health_manager.HealthManager')
@mock.patch('oslo_messaging.Target')
class EngineBasicTest(base.SenlinTestCase):

    def setUp(self):

        super(EngineBasicTest, self).setUp()
        self.eng = service.EngineService('host-a', 'topic-a')
        self.fake_id = '4db0a14c-dc10-4131-8ed6-7573987ce9b0'
        self.gen_id = self.patchobject(uuidutils, 'generate_uuid',
                                       return_value=self.fake_id)

        self.fake_rpc_server = mock.Mock()
        self.get_rpc = self.patchobject(rpc_messaging, 'get_rpc_server',
                                        return_value=self.fake_rpc_server)

    # TODO(Yanyan Hu): Remove this decorator after DB session related
    # work is done.
    @mock.patch.object(context, 'RequestContext')
    def test_engine_start(self, mock_context, mock_msg_cls, mock_hm_cls,
                          mock_disp_cls):

        mock_disp = mock_disp_cls.return_value
        mock_hm = mock_hm_cls.return_value
        mock_target = mock_msg_cls.return_value

        self.eng.start()

        self.gen_id.assert_called_once_with()
        self.assertEqual(self.fake_id, self.eng.engine_id)
        self.assertIsNotNone(self.eng.TG)

        mock_disp_cls.assert_called_once_with(self.eng,
                                              self.eng.dispatcher_topic,
                                              consts.RPC_API_VERSION,
                                              self.eng.TG)
        self.assertEqual(mock_disp, self.eng.dispatcher)
        mock_disp.start.assert_called_once_with()

        mock_hm_cls.assert_called_once_with(self.eng,
                                            self.eng.health_mgr_topic,
                                            consts.RPC_API_VERSION)
        self.assertEqual(mock_hm, self.eng.health_mgr)
        mock_hm.start.assert_called_once_with()

        mock_msg_cls.assert_called_once_with(version=consts.RPC_API_VERSION,
                                             server=self.eng.host,
                                             topic=self.eng.topic)

        self.get_rpc.assert_called_once_with(mock_target, self.eng,
                                             serializer=mock.ANY)

        self.assertEqual(self.fake_rpc_server, self.eng._rpc_server)
        self.fake_rpc_server.start.assert_called_once_with()

    @mock.patch.object(service_obj.Service, 'delete')
    def test_engine_stop(self, mock_delete, mock_msg_cls, mock_hm_cls,
                         mock_disp_cls):
        mock_disp = mock_disp_cls.return_value
        mock_hm = mock_hm_cls.return_value
        self.eng.start()

        self.eng.stop()

        self.fake_rpc_server.stop.assert_called_once_with()
        self.fake_rpc_server.wait.assert_called_once_with()

        mock_disp.stop.assert_called_once_with()
        mock_hm.stop.assert_called_once_with()

        mock_delete.assert_called_once_with(self.fake_id)

    def test_engine_stop_with_exception(self, mock_msg_cls, mock_hm_cls,
                                        mock_disp_cls):
        mock_disp = mock_disp_cls.return_value
        mock_hm = mock_hm_cls.return_value
        self.fake_rpc_server.wait.side_effect = Exception('blah')

        self.eng.start()
        self.eng.stop()

        # on exception, both methods are still invoked
        self.fake_rpc_server.stop.assert_called_once_with()
        self.fake_rpc_server.wait.assert_called_once_with()

        mock_disp.stop.assert_called_once_with()
        mock_hm.stop.assert_called_once_with()


class EngineStatusTest(base.SenlinTestCase):

    def setUp(self):
        super(EngineStatusTest, self).setUp()
        self.eng = service.EngineService('host-a', 'topic-a')
        fake_id = '4db0a14c-dc10-4131-8ed6-7573987ce9b0'
        self.gen_id = self.patchobject(uuidutils, 'generate_uuid',
                                       return_value=fake_id)

        self.fake_rpc_server = mock.Mock()
        self.get_rpc = self.patchobject(rpc_messaging, 'get_rpc_server',
                                        return_value=self.fake_rpc_server)

    @mock.patch('senlin.engine.dispatcher.Dispatcher')
    @mock.patch('senlin.engine.health_manager.HealthManager')
    @mock.patch('oslo_messaging.Target')
    @mock.patch.object(service_obj.Service, 'update')
    def test_engine_manage_report(self, mock_update, mock_msg_cls, mock_hm_cls,
                                  mock_disp_cls):
        cfg.CONF.set_override('periodic_interval', 1)

        # start engine and verify that update is being called more than once
        self.eng.start()
        eventlet.sleep(6)
        self.assertGreater(mock_update.call_count, 1)
        self.eng.stop()

    @mock.patch('senlin.engine.dispatcher.Dispatcher')
    @mock.patch('senlin.engine.health_manager.HealthManager')
    @mock.patch('oslo_messaging.Target')
    @mock.patch.object(service_obj.Service, 'update')
    def test_engine_manage_report_with_exception(self, mock_update,
                                                 mock_msg_cls, mock_hm_cls,
                                                 mock_disp_cls):
        cfg.CONF.set_override('periodic_interval', 1)

        # start engine and verify that update is being called more than once
        # even with the exception being thrown
        mock_update.side_effect = Exception('blah')
        self.eng.start()
        eventlet.sleep(6)
        self.assertGreater(mock_update.call_count, 1)
        self.eng.stop()

    @mock.patch.object(service_obj.Service, 'update')
    def test_service_manage_report_update(self, mock_update):
        mock_update.return_value = mock.Mock()
        self.eng.service_manage_report()
        mock_update.assert_called_once_with(mock.ANY, self.eng.engine_id)

    @mock.patch.object(service_obj.Service, 'gc_by_engine')
    @mock.patch.object(service_obj.Service, 'get_all')
    @mock.patch.object(service_obj.Service, 'delete')
    def test__service_manage_cleanup(self, mock_delete, mock_get_all, mock_gc):
        delta = datetime.timedelta(seconds=2 * cfg.CONF.periodic_interval)
        ages_a_go = timeutils.utcnow(True) - delta
        mock_get_all.return_value = [{'id': 'foo', 'updated_at': ages_a_go}]
        self.eng._service_manage_cleanup()
        mock_delete.assert_called_once_with('foo')
        mock_gc.assert_called_once_with('foo')

    @mock.patch('senlin.engine.dispatcher.Dispatcher')
    @mock.patch('senlin.engine.health_manager.HealthManager')
    @mock.patch('oslo_messaging.Target')
    @mock.patch.object(service_obj.Service, 'get_all')
    def test_service_manage_cleanup(self, mock_get_all, mock_msg_cls,
                                    mock_hm_cls, mock_disp_cls):
        cfg.CONF.set_override('periodic_interval', 1)

        # start engine and verify that get_all is being called more than once
        self.eng.start()
        eventlet.sleep(6)
        self.assertGreater(mock_get_all.call_count, 1)
        self.eng.stop()

    @mock.patch('senlin.engine.dispatcher.Dispatcher')
    @mock.patch('senlin.engine.health_manager.HealthManager')
    @mock.patch('oslo_messaging.Target')
    @mock.patch.object(service_obj.Service, 'get_all')
    def test_service_manage_cleanup_with_exception(self, mock_get_all,
                                                   mock_msg_cls, mock_hm_cls,
                                                   mock_disp_cls):
        cfg.CONF.set_override('periodic_interval', 1)

        # start engine and verify that get_all is being called more than once
        # even with the exception being thrown
        mock_get_all.side_effect = Exception('blah')
        self.eng.start()
        eventlet.sleep(6)
        self.assertGreater(mock_get_all.call_count, 1)
        self.eng.stop()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from sqlalchemy import Boolean, Column, DateTime, Float, ForeignKey
from sqlalchemy import Integer, MetaData, String, Table, Text
from senlin.db.sqlalchemy import types


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine

    profile = Table(
        'profile', meta,
        Column('id', String(36), primary_key=True, nullable=False),
        Column('name', String(255)),
        Column('type', String(255)),
        Column('context', types.Dict),
        Column('spec', types.Dict),
        Column('user', String(32), nullable=False),
        Column('project', String(32), nullable=False),
        Column('domain', String(32)),
        Column('permission', String(32)),
        Column('meta_data', types.Dict),
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    cluster = Table(
        'cluster', meta,
        Column('id', String(36), primary_key=True, nullable=False),
        Column('name', String(255), nullable=False),
        Column('profile_id', String(36), ForeignKey('profile.id'),
               nullable=False),
        Column('user', String(32), nullable=False),
        Column('project', String(32), nullable=False),
        Column('domain', String(32)),
        Column('parent', String(36)),
        Column('init_at', DateTime),
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('min_size', Integer),
        Column('max_size', Integer),
        Column('desired_capacity', Integer),
        Column('next_index', Integer),
        Column('timeout', Integer),
        Column('status', String(255)),
        Column('status_reason', Text),
        Column('meta_data', types.Dict),
        Column('data', types.Dict),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    node = Table(
        'node', meta,
        Column('id', String(36), primary_key=True, nullable=False),
        Column('name', String(255)),
        Column('physical_id', String(36)),
        Column('cluster_id', String(36)),
        Column('profile_id', String(36), ForeignKey('profile.id'),
               nullable=False),
        Column('user', String(32)),
        Column('project', String(32)),
        Column('domain', String(32)),
        Column('index', Integer),
        Column('role', String(64)),
        Column('init_at', DateTime),
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('status', String(255)),
        Column('status_reason', Text),
        Column('meta_data', types.Dict),
        Column('data', types.Dict),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    cluster_lock = Table(
        'cluster_lock', meta,
        Column('cluster_id', String(36), primary_key=True, nullable=False),
        Column('action_ids', types.List),
        Column('semaphore', Integer),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    node_lock = Table(
        'node_lock', meta,
        Column('node_id', String(36), primary_key=True, nullable=False),
        Column('action_id', String(36)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    policy = Table(
        'policy', meta,
        Column('id', String(36), primary_key=True, nullable=False),
        Column('name', String(255)),
        Column('type', String(255)),
        Column('user', String(32), nullable=False),
        Column('project', String(32), nullable=False),
        Column('domain', String(32)),
        Column('cooldown', Integer),
        Column('level', Integer),
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('spec', types.Dict),
        Column('data', types.Dict),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    cluster_policy = Table(
        'cluster_policy', meta,
        Column('id', String(36), primary_key=True, nullable=False),
        Column('cluster_id', String(36), ForeignKey('cluster.id'),
               nullable=False),
        Column('policy_id', String(36), ForeignKey('policy.id'),
               nullable=False),
        Column('cooldown', Integer),
        Column('priority', Integer),
        Column('level', Integer),
        Column('enabled', Boolean),
        Column('data', types.Dict),
        Column('last_op', DateTime),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    receiver = Table(
        'receiver', meta,
        Column('id', String(36), primary_key=True, nullable=False),
        Column('name', String(255)),
        Column('type', String(255)),
        Column('user', String(32)),
        Column('project', String(32)),
        Column('domain', String(32)),
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('cluster_id', String(36)),
        Column('actor', types.Dict),
        Column('action', Text),
        Column('params', types.Dict),
        Column('channel', types.Dict),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    credential = Table(
        'credential', meta,
        Column('user', String(32), primary_key=True, nullable=False),
        Column('project', String(32), primary_key=True, nullable=False),
        Column('cred', types.Dict, nullable=False),
        Column('data', types.Dict),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    action = Table(
        'action', meta,
        Column('id', String(36), primary_key=True, nullable=False),
        Column('name', String(63)),
        Column('context', types.Dict),
        Column('target', String(36)),
        Column('action', Text),
        Column('cause', String(255)),
        Column('owner', String(36)),
        Column('interval', Integer),
        # FIXME: Don't specify fixed precision.
        Column('start_time', Float(precision='24,8')),
        Column('end_time', Float(precision='24,8')),
        Column('timeout', Integer),
        Column('control', String(255)),
        Column('status', String(255)),
        Column('status_reason', Text),
        Column('inputs', types.Dict),
        Column('outputs', types.Dict),
        Column('depends_on', types.List),
        Column('depended_by', types.List),
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('data', types.Dict),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    dependency = Table(
        'dependency', meta,
        Column('id', String(36), nullable=False, primary_key=True),
        Column('depended', String(36), ForeignKey('action.id'),
               nullable=False),
        Column('dependent', String(36), ForeignKey('action.id'),
               nullable=False),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    event = Table(
        'event', meta,
        Column('id', String(36), primary_key=True, nullable=False),
        Column('timestamp', DateTime, nullable=False),
        Column('obj_id', String(36)),
        Column('obj_name', String(255)),
        Column('obj_type', String(36)),
        Column('cluster_id', String(36)),
        Column('level', String(63)),
        Column('user', String(32)),
        Column('project', String(32)),
        Column('action', String(36)),
        Column('status', String(255)),
        Column('status_reason', Text),
        Column('meta_data', types.Dict),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    tables = (
        profile,
        cluster,
        node,
        cluster_lock,
        node_lock,
        policy,
        cluster_policy,
        credential,
        action,
        dependency,
        receiver,
        event,
    )

    for index, table in enumerate(tables):
        try:
            table.create()
        except Exception:
            # If an error occurs, drop all tables created so far to return
            # to the previously existing state.
            meta.drop_all(tables=tables[:index])
            raise


def downgrade(migrate_engine):
    raise NotImplementedError('Database downgrade not supported - '
                              'would drop all tables')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from oslo_config import cfg
from oslo_context import context
import oslo_messaging

from senlin.common import consts
from senlin.common import messaging
from senlin.engine import dispatcher
from senlin.engine import scheduler
from senlin.engine import service
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class TestDispatcher(base.SenlinTestCase):

    def setUp(self):
        super(TestDispatcher, self).setUp()

        self.context = utils.dummy_context()

        self.thm = scheduler.ThreadGroupManager()
        self.svc = service.EngineService('HOST', 'TOPIC')
        self.svc.engine_id = '1234'

    def test_init(self):
        disp = dispatcher.Dispatcher(self.svc, 'TOPIC', '1', self.thm)

        self.assertEqual(self.thm, disp.TG)
        self.assertEqual('1234', disp.engine_id)
        self.assertEqual('TOPIC', disp.topic)
        self.assertEqual('1', disp.version)

    @mock.patch.object(messaging, 'get_rpc_server')
    @mock.patch.object(oslo_messaging, 'Target')
    def test_start(self, mock_target, mock_server):
        mock_server.return_value = mock.Mock()
        mock_target.return_value = mock.Mock()

        disp = dispatcher.Dispatcher(self.svc, 'TOPIC', '1', self.thm)
        disp.start()

        mock_target.assert_called_once_with(server='1234', topic='TOPIC',
                                            version='1')
        the_target = mock_target.return_value
        mock_server.assert_called_once_with(the_target, disp)

        the_server = mock_server.return_value
        the_server.start.assert_called_once_with()

    def test_listening(self):
        disp = dispatcher.Dispatcher(self.svc, 'TOPIC', '1', self.thm)
        result = disp.listening(self.context)
        self.assertTrue(result)

    @mock.patch.object(scheduler.ThreadGroupManager, 'start_action')
    def test_start_action(self, mock_start):
        disp = dispatcher.Dispatcher(self.svc, 'TOPIC', '1', self.thm)
        disp.start_action(self.context, action_id='FOO')

        mock_start.assert_called_once_with('1234', 'FOO')
        mock_start.reset_mock()

        disp.start_action(self.context)
        mock_start.assert_called_once_with('1234', None)

    @mock.patch.object(scheduler.ThreadGroupManager, 'cancel_action')
    def test_cancel_action(self, mock_cancel):
        disp = dispatcher.Dispatcher(self.svc, 'TOPIC', '1', self.thm)
        disp.cancel_action(self.context, action_id='FOO')

        mock_cancel.assert_called_once_with('FOO')

    @mock.patch.object(scheduler.ThreadGroupManager, 'suspend_action')
    def test_suspend_action(self, mock_suspend):
        disp = dispatcher.Dispatcher(self.svc, 'TOPIC', '1', self.thm)
        disp.suspend_action(self.context, action_id='FOO')

        mock_suspend.assert_called_once_with('FOO')

    @mock.patch.object(scheduler.ThreadGroupManager, 'resume_action')
    def test_resume_action(self, mock_resume):
        disp = dispatcher.Dispatcher(self.svc, 'TOPIC', '1', self.thm)
        disp.resume_action(self.context, action_id='FOO')

        mock_resume.assert_called_once_with('FOO')

    @mock.patch.object(scheduler.ThreadGroupManager, 'stop')
    def test_stop(self, mock_stop):
        disp = dispatcher.Dispatcher(self.svc, 'TOPIC', '1', self.thm)
        disp.stop()

        mock_stop.assert_called_once_with(True)

    @mock.patch.object(context, 'get_current')
    @mock.patch.object(messaging, 'get_rpc_client')
    def test_notify_broadcast(self, mock_rpc, mock_get_current):
        cfg.CONF.set_override('host', 'HOSTNAME')
        fake_ctx = mock.Mock()
        mock_get_current.return_value = fake_ctx
        mock_rpc.return_value = mock.Mock()

        dispatcher.notify('METHOD')

        mock_rpc.assert_called_once_with(consts.DISPATCHER_TOPIC, 'HOSTNAME')
        mock_client = mock_rpc.return_value
        mock_client.prepare.assert_called_once_with(fanout=True)

        mock_context = mock_client.prepare.return_value
        mock_context.cast.assert_called_once_with(fake_ctx, 'METHOD')

    @mock.patch.object(context, 'get_current')
    @mock.patch.object(messaging, 'get_rpc_client')
    def test_notify_single_server(self, mock_rpc, mock_get_current):
        cfg.CONF.set_override('host', 'HOSTNAME')
        fake_ctx = mock.Mock()
        mock_get_current.return_value = fake_ctx
        mock_rpc.return_value = mock.Mock()

        result = dispatcher.notify('METHOD', 'FAKE_ENGINE')

        self.assertTrue(result)
        mock_rpc.assert_called_once_with(consts.DISPATCHER_TOPIC, 'HOSTNAME')
        mock_client = mock_rpc.return_value
        mock_client.prepare.assert_called_once_with(server='FAKE_ENGINE')

        mock_context = mock_client.prepare.return_value
        mock_context.cast.assert_called_once_with(fake_ctx, 'METHOD')

    @mock.patch.object(messaging, 'get_rpc_client')
    def test_notify_timeout(self, mock_rpc):
        cfg.CONF.set_override('host', 'HOSTNAME')
        mock_rpc.return_value = mock.Mock()
        mock_client = mock_rpc.return_value
        mock_context = mock_client.prepare.return_value
        mock_context.cast.side_effect = oslo_messaging.MessagingTimeout

        result = dispatcher.notify('METHOD')

        self.assertFalse(result)
        mock_rpc.assert_called_once_with(consts.DISPATCHER_TOPIC, 'HOSTNAME')
        mock_client.prepare.assert_called_once_with(fanout=True)

        mock_context.cast.assert_called_once_with(mock.ANY, 'METHOD')

    @mock.patch.object(dispatcher, 'notify')
    def test_start_action_function(self, mock_notify):
        dispatcher.start_action(engine_id='FAKE_ENGINE')

        mock_notify.assert_called_once_with(dispatcher.START_ACTION,
                                            'FAKE_ENGINE')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_log import log as logging
from senlin import objects

LOG = logging.getLogger(__name__)


class Endpoints(object):

    def __init__(self, project_id, engine_id, recover_action):
        self.engine_id = engine_id
        self.project_id = project_id
        self.recover_action = recover_action

    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        raise NotImplementedError

    def _check_registry_status(self, ctx, engine_id, cluster_id):
        registry = objects.HealthRegistry.get_by_engine(ctx, engine_id,
                                                        cluster_id)

        if registry is None:
            return False

        if registry.enabled is True:
            return True

        return False



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from oslo_messaging.rpc import dispatcher as rpc
import six

from senlin.common import consts
from senlin.common import exception as exc
from senlin.engine.actions import base as am
from senlin.engine import cluster as cm
from senlin.engine import dispatcher
from senlin.engine import service
from senlin.objects import cluster as co
from senlin.objects import node as no
from senlin.objects.requests import clusters as orco
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class ClusterOpTest(base.SenlinTestCase):

    def setUp(self):
        super(ClusterOpTest, self).setUp()

        self.ctx = utils.dummy_context(project='cluster_op_test_project')
        self.eng = service.EngineService('host-a', 'topic-a')

    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(no.Node, 'ids_by_cluster')
    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_op(self, mock_find, mock_cluster, mock_nodes, mock_action,
                        mock_start):
        x_db_cluster = mock.Mock()
        mock_find.return_value = x_db_cluster
        x_schema = mock.Mock()
        x_profile = mock.Mock(OPERATIONS={'dance': x_schema})
        x_cluster = mock.Mock(id='12345678AB')
        x_cluster.rt = {'profile': x_profile}
        mock_cluster.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        params = {'style': 'tango'}
        filters = {'role': 'slave'}
        mock_nodes.return_value = ['NODE1', 'NODE2']
        req = orco.ClusterOperationRequest(identity='FAKE_CLUSTER',
                                           operation='dance',
                                           params=params,
                                           filters=filters)

        result = self.eng.cluster_op(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')
        mock_cluster.assert_called_once_with(self.ctx, dbcluster=x_db_cluster)
        x_schema.validate.assert_called_once_with({'style': 'tango'})
        mock_nodes.assert_called_once_with(self.ctx, '12345678AB',
                                           filters={'role': 'slave'})
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', consts.CLUSTER_OPERATION,
            name='cluster_dance_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={
                'operation': 'dance',
                'params': {'style': 'tango'},
                'nodes': ['NODE1', 'NODE2']
            }
        )
        mock_start.assert_called_once_with()

    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_op_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(
            type='cluster', id='Bogus')
        req = orco.ClusterOperationRequest(identity='Bogus', operation='dance')
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_op,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual("The cluster 'Bogus' could not be found.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')

    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_op_unsupported_operation(self, mock_find, mock_cluster):
        x_db_cluster = mock.Mock(id='12345678AB')
        mock_find.return_value = x_db_cluster
        x_schema = mock.Mock()
        x_profile = mock.Mock(OPERATIONS={'dance': x_schema}, type='cow')
        x_cluster = mock.Mock()
        x_cluster.rt = {'profile': x_profile}
        mock_cluster.return_value = x_cluster
        req = orco.ClusterOperationRequest(identity='node1', operation='swim')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_op,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The requested operation 'swim' is not supported "
                         "by the profile type 'cow'.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'node1')
        mock_cluster.assert_called_once_with(self.ctx, dbcluster=x_db_cluster)

    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_op_bad_parameters(self, mock_find, mock_cluster):
        x_db_cluster = mock.Mock(id='12345678AB')
        mock_find.return_value = x_db_cluster
        x_schema = mock.Mock()
        x_schema.validate.side_effect = exc.ESchema(message='Boom')
        x_profile = mock.Mock(OPERATIONS={'dance': x_schema})
        x_cluster = mock.Mock()
        x_cluster.rt = {'profile': x_profile}
        mock_cluster.return_value = x_cluster
        req = orco.ClusterOperationRequest(identity='node1', operation='dance',
                                           params={'style': 'tango'})

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_op,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Boom.", six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'node1')
        mock_cluster.assert_called_once_with(self.ctx, dbcluster=x_db_cluster)
        x_schema.validate.assert_called_once_with({'style': 'tango'})

    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(no.Node, 'ids_by_cluster')
    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_op_no_parameters(self, mock_find, mock_cluster,
                                      mock_nodes, mock_action, mock_start):
        x_db_cluster = mock.Mock()
        mock_find.return_value = x_db_cluster
        x_schema = mock.Mock()
        x_profile = mock.Mock(OPERATIONS={'dance': x_schema})
        x_cluster = mock.Mock(id='12345678AB')
        x_cluster.rt = {'profile': x_profile}
        mock_cluster.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        filters = {'role': 'slave'}
        mock_nodes.return_value = ['NODE1', 'NODE2']
        req = orco.ClusterOperationRequest(identity='FAKE_CLUSTER',
                                           operation='dance',
                                           filters=filters)

        result = self.eng.cluster_op(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')
        mock_cluster.assert_called_once_with(self.ctx, dbcluster=x_db_cluster)
        self.assertEqual(0, x_schema.validate.call_count)
        mock_nodes.assert_called_once_with(self.ctx, '12345678AB',
                                           filters={'role': 'slave'})
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', consts.CLUSTER_OPERATION,
            name='cluster_dance_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={
                'operation': 'dance',
                'params': {},
                'nodes': ['NODE1', 'NODE2']
            }
        )
        mock_start.assert_called_once_with()

    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(no.Node, 'ids_by_cluster')
    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_op_no_filters(self, mock_find, mock_cluster,
                                   mock_nodes, mock_action, mock_start):
        x_db_cluster = mock.Mock()
        mock_find.return_value = x_db_cluster
        x_schema = mock.Mock()
        x_profile = mock.Mock(OPERATIONS={'dance': x_schema})
        x_cluster = mock.Mock(id='12345678AB')
        x_cluster.rt = {'profile': x_profile}
        mock_cluster.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        mock_nodes.return_value = ['NODE1', 'NODE2']
        req = orco.ClusterOperationRequest(identity='FAKE_CLUSTER',
                                           operation='dance')

        result = self.eng.cluster_op(self.ctx, req.obj_to_primitive())

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')
        mock_cluster.assert_called_once_with(self.ctx, dbcluster=x_db_cluster)
        self.assertEqual(0, x_schema.validate.call_count)
        mock_nodes.assert_called_once_with(self.ctx, '12345678AB')
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', consts.CLUSTER_OPERATION,
            name='cluster_dance_12345678',
            cause=consts.CAUSE_RPC,
            status=am.Action.READY,
            inputs={
                'operation': 'dance',
                'params': {},
                'nodes': ['NODE1', 'NODE2']
            }
        )
        mock_start.assert_called_once_with()

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(no.Node, 'ids_by_cluster')
    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_op_bad_filters(self, mock_find, mock_cluster,
                                    mock_nodes, mock_action):
        x_db_cluster = mock.Mock()
        mock_find.return_value = x_db_cluster
        x_schema = mock.Mock()
        x_profile = mock.Mock(OPERATIONS={'dance': x_schema})
        x_cluster = mock.Mock(id='12345678AB')
        x_cluster.rt = {'profile': x_profile}
        mock_cluster.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        mock_nodes.return_value = ['NODE1', 'NODE2']
        filters = {'shape': 'round'}
        req = orco.ClusterOperationRequest(identity='FAKE_CLUSTER',
                                           operation='dance',
                                           filters=filters)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_op,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("Filter key 'shape' is unsupported.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')
        mock_cluster.assert_called_once_with(self.ctx, dbcluster=x_db_cluster)
        self.assertEqual(0, x_schema.validate.call_count)
        self.assertEqual(0, mock_nodes.call_count)
        self.assertEqual(0, mock_action.call_count)

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(no.Node, 'ids_by_cluster')
    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(co.Cluster, 'find')
    def test_cluster_op_no_nodes_found(self, mock_find, mock_cluster,
                                       mock_nodes, mock_action):
        x_db_cluster = mock.Mock()
        mock_find.return_value = x_db_cluster
        x_schema = mock.Mock()
        x_profile = mock.Mock(OPERATIONS={'dance': x_schema})
        x_cluster = mock.Mock(id='12345678AB')
        x_cluster.rt = {'profile': x_profile}
        mock_cluster.return_value = x_cluster
        mock_nodes.return_value = []
        mock_action.return_value = 'ACTION_ID'
        filters = {'role': 'slave'}
        req = orco.ClusterOperationRequest(identity='FAKE_CLUSTER',
                                           operation='dance', filters=filters)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_op,
                               self.ctx, req.obj_to_primitive())

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("No node (matching the filter) could be found.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')
        mock_cluster.assert_called_once_with(self.ctx, dbcluster=x_db_cluster)
        mock_nodes.assert_called_once_with(self.ctx, '12345678AB',
                                           filters={'role': 'slave'})
        self.assertEqual(0, mock_action.call_count)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import testtools

from senlin.common import consts
from senlin.events import base
from senlin.tests.unit.common import utils

CLUSTER_ID = '2c5139a6-24ba-4a6f-bd53-a268f61536de'


class TestEventBackend(testtools.TestCase):

    def setUp(self):
        super(TestEventBackend, self).setUp()
        self.ctx = utils.dummy_context()

    @mock.patch('oslo_utils.reflection.get_class_name')
    def test__check_entity_cluster(self, mock_get):
        entity = mock.Mock()
        mock_get.return_value = 'Cluster'

        res = base.EventBackend._check_entity(entity)

        self.assertEqual('CLUSTER', res)
        mock_get.assert_called_once_with(entity, fully_qualified=False)

    @mock.patch('oslo_utils.reflection.get_class_name')
    def test__check_entity_node(self, mock_get):
        entity = mock.Mock()
        mock_get.return_value = 'Node'

        res = base.EventBackend._check_entity(entity)

        self.assertEqual('NODE', res)
        mock_get.assert_called_once_with(entity, fully_qualified=False)

    def test__get_action_name_unexpected(self):
        action = mock.Mock(action="UNEXPECTED")
        res = base.EventBackend._get_action_name(action)
        self.assertEqual('unexpected', res)

    def test__get_action_name_correct_format(self):
        action = mock.Mock(action="FOO_BAR")
        res = base.EventBackend._get_action_name(action)
        self.assertEqual('bar', res)

    def test__get_action_name_operation_found(self):
        action = mock.Mock(action=consts.NODE_OPERATION,
                           inputs={'operation': 'bar'})
        res = base.EventBackend._get_action_name(action)
        self.assertEqual('bar', res)

    def test__get_action_name_operation_not_found(self):
        action = mock.Mock(action="FOO_OPERATION", inputs={})
        res = base.EventBackend._get_action_name(action)
        self.assertEqual('operation', res)

    def test_dump(self):
        self.assertRaises(NotImplementedError,
                          base.EventBackend.dump,
                          '1', '2')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlin.drivers import base
from senlin.drivers import sdk


class GlanceClient(base.DriverBase):
    """Glance V2 driver."""

    def __init__(self, params):
        super(GlanceClient, self).__init__(params)
        self.conn = sdk.create_connection(params)
        self.session = self.conn.session

    @sdk.translate_exception
    def image_find(self, name_or_id, ignore_missing=True):
        return self.conn.image.find_image(name_or_id, ignore_missing)

    @sdk.translate_exception
    def image_get(self, image):
        return self.conn.image.get_image(image)

    @sdk.translate_exception
    def image_delete(self, name_or_id, ignore_missing=False):
        return self.conn.image.delete_image(name_or_id, ignore_missing)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils
import six
from stevedore import named

from senlin.common import consts

LOG = logging.getLogger(__name__)
FMT = '%(name)s[%(obj_id)s] %(action)s[%(id)s] %(phase)s: %(reason)s'
dispatchers = None


def load_dispatcher():
    """Load dispatchers."""
    global dispatchers

    LOG.debug("Loading dispatchers")
    dispatchers = named.NamedExtensionManager(
        namespace="senlin.dispatchers",
        names=cfg.CONF.event_dispatchers,
        invoke_on_load=True,
        propagate_map_exceptions=True)
    if not list(dispatchers):
        LOG.warning("No dispatchers configured for 'senlin.dispatchers'")
    else:
        LOG.info("Loaded dispatchers: %s", dispatchers.names())


def _event_data(action, phase=None, reason=None):
    action_name = action.action
    if action_name in [consts.NODE_OPERATION, consts.CLUSTER_OPERATION]:
        action_name = action.inputs.get('operation', action_name)
    name = action.entity.name if action.entity else "Unknown"
    obj_id = action.entity.id[:8] if action.entity else "Unknown"

    return dict(name=name,
                obj_id=obj_id,
                action=action_name,
                id=action.id[:8],
                phase=phase,
                reason=reason)


def _dump(level, action, phase, reason, timestamp):
    global dispatchers

    if timestamp is None:
        timestamp = timeutils.utcnow(True)

    # We check the logging level threshold only when debug is False
    if cfg.CONF.debug is False:
        watermark = cfg.CONF.dispatchers.priority.upper()
        bound = consts.EVENT_LEVELS.get(watermark, logging.INFO)
        if level < bound:
            return

    if cfg.CONF.dispatchers.exclude_derived_actions:
        if action.cause == consts.CAUSE_DERIVED:
            return

    try:
        dispatchers.map_method("dump", level, action,
                               phase=phase, reason=reason, timestamp=timestamp)
    except Exception as ex:
        LOG.exception("Dispatcher failed to handle the event: %s",
                      six.text_type(ex))


def critical(action, phase=None, reason=None, timestamp=None):
    _dump(logging.CRITICAL, action, phase, reason, timestamp)
    LOG.critical(FMT, _event_data(action, phase, reason))


def error(action, phase=None, reason=None, timestamp=None):
    _dump(logging.ERROR, action, phase, reason, timestamp)
    LOG.error(FMT, _event_data(action, phase, reason))


def warning(action, phase=None, reason=None, timestamp=None):
    _dump(logging.WARNING, action, phase, reason, timestamp)
    LOG.warning(FMT, _event_data(action, phase, reason))


def info(action, phase=None, reason=None, timestamp=None):
    _dump(logging.INFO, action, phase, reason, timestamp)
    LOG.info(FMT, _event_data(action, phase, reason))


def debug(action, phase=None, reason=None, timestamp=None):
    _dump(logging.DEBUG, action, phase, reason, timestamp)
    LOG.debug(FMT, _event_data(action, phase, reason))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
ClusterPolicies endpoint for Senlin v1 REST API.
"""

from webob import exc

from senlin.api.common import util
from senlin.api.common import wsgi
from senlin.common import consts
from senlin.common.i18n import _


class ClusterPolicyController(wsgi.Controller):
    """WSGI controller for Cluster-Policy binding in Senlin v1 API."""

    # Define request scope
    # (must match what is in policy file and policies in code.)
    REQUEST_SCOPE = 'cluster_policies'

    @util.policy_enforce
    def index(self, req, cluster_id):
        param_whitelist = {
            consts.CP_ENABLED: 'single',
            consts.CP_POLICY_NAME: 'single',
            consts.CP_POLICY_TYPE: 'single',
            consts.PARAM_SORT: 'single',
        }
        for key in req.params.keys():
            if (key not in param_whitelist.keys()):
                raise exc.HTTPBadRequest(_('Invalid parameter %s') % key)

        params = util.get_allowed_params(req.params, param_whitelist)
        key = consts.CP_ENABLED
        if key in params:
            params[key] = util.parse_bool_param(key, params[key])
        params['identity'] = cluster_id

        obj = util.parse_request('ClusterPolicyListRequest', req, params)
        policies = self.rpc_client.call(req.context, 'cluster_policy_list',
                                        obj)

        return {'cluster_policies': policies}

    @util.policy_enforce
    def get(self, req, cluster_id, policy_id):

        params = {'identity': cluster_id, 'policy_id': policy_id}
        obj = util.parse_request('ClusterPolicyGetRequest', req, params)
        cluster_policy = self.rpc_client.call(req.context,
                                              'cluster_policy_get', obj)
        return {'cluster_policy': cluster_policy}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
SQLAlchemy models for Senlin data.
"""

from oslo_db.sqlalchemy import models
from oslo_utils import uuidutils
from sqlalchemy import Boolean, Column, Numeric, ForeignKey, Integer
from sqlalchemy import String, Text
from sqlalchemy.ext import declarative
from sqlalchemy.orm import backref
from sqlalchemy.orm import relationship

from senlin.db.sqlalchemy import types

BASE = declarative.declarative_base()
UUID4 = uuidutils.generate_uuid


class TimestampMixin(object):
    created_at = Column(types.TZAwareDateTime)
    updated_at = Column(types.TZAwareDateTime)


class Profile(BASE, TimestampMixin, models.ModelBase):
    """Profile objects."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __tablename__ = 'profile'

    id = Column('id', String(36), primary_key=True, default=lambda: UUID4())
    name = Column(String(255))
    type = Column(String(255))
    context = Column(types.Dict)
    spec = Column(types.Dict)
    user = Column(String(32), nullable=False)
    project = Column(String(32), nullable=False)
    domain = Column(String(32))
    permission = Column(String(32))
    meta_data = Column(types.Dict)


class Policy(BASE, TimestampMixin, models.ModelBase):
    """Policy objects."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __tablename__ = 'policy'

    id = Column('id', String(36), primary_key=True, default=lambda: UUID4())
    user = Column(String(32), nullable=False)
    project = Column(String(32), nullable=False)
    domain = Column(String(32))
    name = Column(String(255))
    type = Column(String(255))
    cooldown = Column(Integer)
    level = Column(Integer)
    spec = Column(types.Dict)
    data = Column(types.Dict)


class Cluster(BASE, TimestampMixin, models.ModelBase):
    """Cluster objects."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __tablename__ = 'cluster'

    id = Column('id', String(36), primary_key=True, default=lambda: UUID4())
    name = Column('name', String(255))
    profile_id = Column(String(36), ForeignKey('profile.id'), nullable=False)
    user = Column(String(32), nullable=False)
    project = Column(String(32), nullable=False)
    domain = Column(String(32))
    parent = Column(String(36))

    init_at = Column(types.TZAwareDateTime)

    min_size = Column(Integer)
    max_size = Column(Integer)
    desired_capacity = Column(Integer)
    next_index = Column(Integer)
    timeout = Column(Integer)

    status = Column(String(255))
    status_reason = Column(Text)
    meta_data = Column(types.Dict)
    data = Column(types.Dict)
    dependents = Column(types.Dict)
    config = Column(types.Dict)


class Node(BASE, TimestampMixin, models.ModelBase):
    """Node objects."""

    __table_args__ = {'mysql_engine': 'InnoDB'}
    __tablename__ = 'node'

    id = Column('id', String(36), primary_key=True, default=lambda: UUID4())
    name = Column(String(255))
    physical_id = Column(String(36))
    cluster_id = Column(String(36))
    profile_id = Column(String(36), ForeignKey('profile.id'))
    user = Column(String(32), nullable=False)
    project = Column(String(32), nullable=False)
    domain = Column(String(32))
    index = Column(Integer)
    role = Column(String(64))

    init_at = Column(types.TZAwareDateTime)

    status = Column(String(255))
    status_reason = Column(Text)
    meta_data = Column(types.Dict)
    data = Column(types.Dict)
    dependents = Column(types.Dict)
    profile = relationship(Profile, backref=backref('nodes'))


class ClusterLock(BASE, models.ModelBase):
    """Cluster locks for actions."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __tablename__ = 'cluster_lock'

    cluster_id = Column(String(36), primary_key=True, nullable=False)
    action_ids = Column(types.List)
    semaphore = Column(Integer)


class NodeLock(BASE, models.ModelBase):
    """Node locks for actions."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __tablename__ = 'node_lock'

    node_id = Column(String(36), primary_key=True, nullable=False)
    action_id = Column(String(36))


class ClusterPolicies(BASE, models.ModelBase):
    """Association between clusters and policies."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __tablename__ = 'cluster_policy'

    id = Column('id', String(36), primary_key=True, default=lambda: UUID4())
    cluster_id = Column(String(36), ForeignKey('cluster.id'), nullable=False)
    policy_id = Column(String(36), ForeignKey('policy.id'), nullable=False)
    cluster = relationship(Cluster, backref=backref('policies'))
    policy = relationship(Policy, backref=backref('bindings'))
    enabled = Column(Boolean)
    priority = Column(Integer)
    data = Column(types.Dict)
    last_op = Column(types.TZAwareDateTime)


class HealthRegistry(BASE, models.ModelBase):
    """Clusters registered for health management."""

    __table_args__ = {'mysql_engine': 'InnoDB'}
    __tablename__ = 'health_registry'

    id = Column('id', String(36), primary_key=True, default=lambda: UUID4())
    cluster_id = Column(String(36), ForeignKey('cluster.id'), nullable=False)
    check_type = Column('check_type', String(255))
    interval = Column(Integer)
    params = Column(types.Dict)
    enabled = Column(Boolean)
    engine_id = Column('engine_id', String(36))


class Receiver(BASE, TimestampMixin, models.ModelBase):
    """Receiver objects associated with clusters."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __tablename__ = 'receiver'

    id = Column('id', String(36), primary_key=True, default=lambda: UUID4())
    name = Column('name', String(255))
    type = Column(String(255))
    user = Column(String(32))
    project = Column(String(32))
    domain = Column(String(32))

    cluster_id = Column(String(36), ForeignKey('cluster.id'))
    actor = Column(types.Dict)
    action = Column(Text)
    params = Column(types.Dict)
    channel = Column(types.Dict)


class Credential(BASE, models.ModelBase):
    """User credentials for keystone trusts etc."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __tablename__ = 'credential'

    user = Column(String(32), primary_key=True, nullable=False)
    project = Column(String(32), primary_key=True, nullable=False)
    cred = Column(types.Dict, nullable=False)
    data = Column(types.Dict)


class ActionDependency(BASE, models.ModelBase):
    """Action dependencies."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __tablename__ = 'dependency'

    id = Column('id', String(36), primary_key=True, default=lambda: UUID4())
    depended = Column('depended', String(36), ForeignKey('action.id'),
                      nullable=False)
    dependent = Column('dependent', String(36), ForeignKey('action.id'),
                       nullable=False)


class Action(BASE, TimestampMixin, models.ModelBase):
    """Action objects."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __tablename__ = 'action'

    id = Column('id', String(36), primary_key=True, default=lambda: UUID4())
    name = Column(String(63))
    context = Column(types.Dict)
    target = Column(String(36))
    action = Column(Text)
    cause = Column(String(255))
    owner = Column(String(36))
    interval = Column(Integer)
    start_time = Column(Numeric(18, 6))
    end_time = Column(Numeric(18, 6))
    timeout = Column(Integer)
    status = Column(String(255))
    status_reason = Column(Text)
    control = Column(String(255))
    inputs = Column(types.Dict)
    outputs = Column(types.Dict)
    data = Column(types.Dict)
    user = Column(String(32))
    project = Column(String(32))
    domain = Column(String(32))


class Event(BASE, models.ModelBase):
    """Events generated by the Senin engine."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __tablename__ = 'event'

    id = Column('id', String(36), primary_key=True, default=lambda: UUID4())
    timestamp = Column(types.TZAwareDateTime)
    oid = Column(String(36))
    oname = Column(String(255))
    otype = Column(String(36))
    cluster_id = Column(String(36), ForeignKey('cluster.id'), nullable=True)
    cluster = relationship(Cluster, backref=backref('events'))
    level = Column(String(64))
    user = Column(String(32))
    project = Column(String(32))
    action = Column(String(36))
    status = Column(String(255))
    status_reason = Column(Text)
    meta_data = Column(types.Dict)

    def as_dict(self):
        data = super(Event, self)._as_dict()
        ts = data['timestamp'].replace(microsecond=0).isoformat()
        data['timestamp'] = ts
        return data


class Service(BASE, TimestampMixin, models.ModelBase):
    """Senlin service engine registry."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __tablename__ = 'service'

    id = Column('id', String(36), primary_key=True, nullable=False)
    host = Column(String(255))
    binary = Column(String(255))
    topic = Column(String(255))
    disabled = Column(Boolean, default=False)
    disabled_reason = Column(String(255))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Action object."""

from oslo_utils import uuidutils

from senlin.common import exception
from senlin.common import utils
from senlin.db import api as db_api
from senlin.objects import base
from senlin.objects import dependency as dobj
from senlin.objects import fields


@base.SenlinObjectRegistry.register
class Action(base.SenlinObject, base.VersionedObjectDictCompat):
    """Senlin action object."""

    fields = {
        'id': fields.UUIDField(),
        'created_at': fields.DateTimeField(),
        'updated_at': fields.DateTimeField(nullable=True),
        'name': fields.StringField(),
        'context': fields.JsonField(),
        'target': fields.UUIDField(),
        'action': fields.StringField(),
        'cause': fields.StringField(),
        'owner': fields.UUIDField(nullable=True),
        'interval': fields.IntegerField(nullable=True),
        'start_time': fields.FloatField(nullable=True),
        'end_time': fields.FloatField(nullable=True),
        'timeout': fields.IntegerField(nullable=True),
        'status': fields.StringField(),
        'status_reason': fields.StringField(nullable=True),
        'control': fields.StringField(nullable=True),
        'inputs': fields.JsonField(nullable=True),
        'outputs': fields.JsonField(nullable=True),
        'data': fields.JsonField(nullable=True),
        'user': fields.StringField(),
        'project': fields.StringField(),
        'domain': fields.StringField(nullable=True),
    }

    @classmethod
    def create(cls, context, values):
        obj = db_api.action_create(context, values)
        return cls._from_db_object(context, cls(context), obj)

    @classmethod
    def find(cls, context, identity, **kwargs):
        """Find an action with the given identity.

        :param context: An instance of the request context.
        :param identity: The UUID, name or short-id of an action.
        :param dict kwargs: Other query parameters.
        :return: A DB object of action or an exception `ResourceNotFound` if
                 no matching action is found.
        """
        if uuidutils.is_uuid_like(identity):
            action = cls.get(context, identity, **kwargs)
            if not action:
                action = cls.get_by_name(context, identity, **kwargs)
        else:
            action = cls.get_by_name(context, identity, **kwargs)
            if not action:
                action = cls.get_by_short_id(context, identity, **kwargs)

        if not action:
            raise exception.ResourceNotFound(type='action', id=identity)

        return action

    @classmethod
    def get(cls, context, action_id, **kwargs):
        obj = db_api.action_get(context, action_id, **kwargs)
        return cls._from_db_object(context, cls(), obj)

    @classmethod
    def get_by_name(cls, context, name, **kwargs):
        obj = db_api.action_get_by_name(context, name, **kwargs)
        return cls._from_db_object(context, cls(), obj)

    @classmethod
    def get_by_short_id(cls, context, short_id, **kwargs):
        obj = db_api.action_get_by_short_id(context, short_id, **kwargs)
        return cls._from_db_object(context, cls(), obj)

    @classmethod
    def get_all(cls, context, **kwargs):
        objs = db_api.action_get_all(context, **kwargs)
        return [cls._from_db_object(context, cls(), obj) for obj in objs]

    @classmethod
    def get_all_by_owner(cls, context, owner):
        objs = db_api.action_get_all_by_owner(context, owner)
        return [cls._from_db_object(context, cls(), obj) for obj in objs]

    @classmethod
    def check_status(cls, context, action_id, timestamp):
        return db_api.action_check_status(context, action_id, timestamp)

    @classmethod
    def mark_succeeded(cls, context, action_id, timestamp):
        return db_api.action_mark_succeeded(context, action_id, timestamp)

    @classmethod
    def mark_ready(cls, context, action_id, timestamp):
        return db_api.action_mark_ready(context, action_id, timestamp)

    @classmethod
    def mark_failed(cls, context, action_id, timestamp, reason=None):
        return db_api.action_mark_failed(context, action_id, timestamp, reason)

    @classmethod
    def mark_cancelled(cls, context, action_id, timestamp):
        return db_api.action_mark_cancelled(context, action_id, timestamp)

    @classmethod
    def acquire(cls, context, action_id, owner, timestamp):
        return db_api.action_acquire(context, action_id, owner, timestamp)

    @classmethod
    def acquire_random_ready(cls, context, owner, timestamp):
        return db_api.action_acquire_random_ready(context, owner, timestamp)

    @classmethod
    def acquire_first_ready(cls, context, owner, timestamp):
        return db_api.action_acquire_first_ready(context, owner, timestamp)

    @classmethod
    def abandon(cls, context, action_id, values=None):
        return db_api.action_abandon(context, action_id, values)

    @classmethod
    def signal(cls, context, action_id, value):
        return db_api.action_signal(context, action_id, value)

    @classmethod
    def signal_query(cls, context, action_id):
        return db_api.action_signal_query(context, action_id)

    @classmethod
    def lock_check(cls, context, action_id, owner=None):
        return db_api.action_lock_check(context, action_id, owner)

    @classmethod
    def update(cls, context, action_id, values):
        return db_api.action_update(context, action_id, values)

    @classmethod
    def delete(cls, context, action_id):
        db_api.action_delete(context, action_id)

    @classmethod
    def delete_by_target(cls, context, target, action=None,
                         action_excluded=None, status=None):
        """Delete an action with the target and other given params.

        :param target: The ID of the target cluster/node
        :param action: A list of actions to be included.
        :param action_excluded: A list of actions to be excluded.
        :param status: A list of statuses to be delete filtered.
        :return: None.
        """
        return db_api.action_delete_by_target(context, target, action=action,
                                              action_excluded=action_excluded,
                                              status=status)

    def to_dict(self):
        if self.id:
            dep_on = dobj.Dependency.get_depended(self.context, self.id)
            dep_by = dobj.Dependency.get_dependents(self.context, self.id)
        else:
            dep_on = []
            dep_by = []
        action_dict = {
            'id': self.id,
            'name': self.name,
            'action': self.action,
            'target': self.target,
            'cause': self.cause,
            'owner': self.owner,
            'interval': self.interval,
            'start_time': self.start_time,
            'end_time': self.end_time,
            'timeout': self.timeout,
            'status': self.status,
            'status_reason': self.status_reason,
            'inputs': self.inputs,
            'outputs': self.outputs,
            'depends_on': dep_on,
            'depended_by': dep_by,
            'created_at': utils.isotime(self.created_at),
            'updated_at': utils.isotime(self.updated_at),
            'data': self.data,
            'user': self.user,
            'project': self.project,
        }
        return action_dict



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Interface for database access.

SQLAlchemy is currently the only supported backend.
"""

from oslo_config import cfg
from oslo_db import api

CONF = cfg.CONF


_BACKEND_MAPPING = {'sqlalchemy': 'senlin.db.sqlalchemy.api'}

IMPL = api.DBAPI.from_config(CONF, backend_mapping=_BACKEND_MAPPING)


def get_engine():
    return IMPL.get_engine()


# Clusters
def cluster_create(context, values):
    return IMPL.cluster_create(context, values)


def cluster_get(context, cluster_id, project_safe=True):
    return IMPL.cluster_get(context, cluster_id, project_safe=project_safe)


def cluster_get_by_name(context, cluster_name, project_safe=True):
    return IMPL.cluster_get_by_name(context, cluster_name,
                                    project_safe=project_safe)


def cluster_get_by_short_id(context, short_id, project_safe=True):
    return IMPL.cluster_get_by_short_id(context, short_id,
                                        project_safe=project_safe)


def cluster_get_all(context, limit=None, marker=None, sort=None, filters=None,
                    project_safe=True):
    return IMPL.cluster_get_all(context, limit=limit, marker=marker, sort=sort,
                                filters=filters, project_safe=project_safe)


def cluster_next_index(context, cluster_id):
    return IMPL.cluster_next_index(context, cluster_id)


def cluster_count_all(context, filters=None, project_safe=True):
    return IMPL.cluster_count_all(context, filters=filters,
                                  project_safe=project_safe)


def cluster_update(context, cluster_id, values):
    return IMPL.cluster_update(context, cluster_id, values)


def cluster_delete(context, cluster_id):
    return IMPL.cluster_delete(context, cluster_id)


# Nodes
def node_create(context, values):
    return IMPL.node_create(context, values)


def node_get(context, node_id, project_safe=True):
    return IMPL.node_get(context, node_id, project_safe=project_safe)


def node_get_by_name(context, name, project_safe=True):
    return IMPL.node_get_by_name(context, name, project_safe=project_safe)


def node_get_by_short_id(context, short_id, project_safe=True):
    return IMPL.node_get_by_short_id(context, short_id,
                                     project_safe=project_safe)


def node_get_all(context, cluster_id=None, limit=None, marker=None, sort=None,
                 filters=None, project_safe=True):
    return IMPL.node_get_all(context, cluster_id=cluster_id, filters=filters,
                             limit=limit, marker=marker, sort=sort,
                             project_safe=project_safe)


def node_get_all_by_cluster(context, cluster_id, filters=None,
                            project_safe=True):
    return IMPL.node_get_all_by_cluster(context, cluster_id, filters=filters,
                                        project_safe=project_safe)


def node_ids_by_cluster(context, cluster_id, filters=None):
    return IMPL.node_ids_by_cluster(context, cluster_id, filters=None)


def node_count_by_cluster(context, cluster_id, **kwargs):
    return IMPL.node_count_by_cluster(context, cluster_id, **kwargs)


def node_update(context, node_id, values):
    return IMPL.node_update(context, node_id, values)


def node_migrate(context, node_id, to_cluster, timestamp, role=None):
    return IMPL.node_migrate(context, node_id, to_cluster, timestamp, role)


def node_delete(context, node_id):
    return IMPL.node_delete(context, node_id)


# Locks
def cluster_lock_acquire(cluster_id, action_id, scope):
    return IMPL.cluster_lock_acquire(cluster_id, action_id, scope)


def cluster_lock_release(cluster_id, action_id, scope):
    return IMPL.cluster_lock_release(cluster_id, action_id, scope)


def cluster_lock_steal(node_id, action_id):
    return IMPL.cluster_lock_steal(node_id, action_id)


def node_lock_acquire(node_id, action_id):
    return IMPL.node_lock_acquire(node_id, action_id)


def node_lock_release(node_id, action_id):
    return IMPL.node_lock_release(node_id, action_id)


def node_lock_steal(node_id, action_id):
    return IMPL.node_lock_steal(node_id, action_id)


# Policies
def policy_create(context, values):
    return IMPL.policy_create(context, values)


def policy_get(context, policy_id, project_safe=True):
    return IMPL.policy_get(context, policy_id, project_safe=project_safe)


def policy_get_by_name(context, name, project_safe=True):
    return IMPL.policy_get_by_name(context, name, project_safe=project_safe)


def policy_get_by_short_id(context, short_id, project_safe=True):
    return IMPL.policy_get_by_short_id(context, short_id,
                                       project_safe=project_safe)


def policy_get_all(context, limit=None, marker=None, sort=None, filters=None,
                   project_safe=True):
    return IMPL.policy_get_all(context, limit=limit, marker=marker, sort=sort,
                               filters=filters, project_safe=project_safe)


def policy_update(context, policy_id, values):
    return IMPL.policy_update(context, policy_id, values)


def policy_delete(context, policy_id):
    return IMPL.policy_delete(context, policy_id)


# Cluster-Policy Associations
def cluster_policy_get(context, cluster_id, policy_id):
    return IMPL.cluster_policy_get(context, cluster_id, policy_id)


def cluster_policy_get_all(context, cluster_id, filters=None, sort=None):
    return IMPL.cluster_policy_get_all(context, cluster_id, filters=filters,
                                       sort=sort)


def cluster_policy_ids_by_cluster(context, cluster_id):
    return IMPL.cluster_policy_ids_by_cluster(context, cluster_id)


def cluster_policy_get_by_type(context, cluster_id, policy_type, filters=None):
    return IMPL.cluster_policy_get_by_type(context, cluster_id, policy_type,
                                           filters=filters)


def cluster_policy_get_by_name(context, cluster_id, policy_name, filters=None):
    return IMPL.cluster_policy_get_by_name(context, cluster_id, policy_name,
                                           filters=filters)


def cluster_policy_attach(context, cluster_id, policy_id, values):
    return IMPL.cluster_policy_attach(context, cluster_id, policy_id, values)


def cluster_policy_detach(context, cluster_id, policy_id):
    return IMPL.cluster_policy_detach(context, cluster_id, policy_id)


def cluster_policy_update(context, cluster_id, policy_id, values):
    return IMPL.cluster_policy_update(context, cluster_id, policy_id, values)


# Profiles
def profile_create(context, values):
    return IMPL.profile_create(context, values)


def profile_get(context, profile_id, project_safe=True):
    return IMPL.profile_get(context, profile_id, project_safe=project_safe)


def profile_get_by_name(context, name, project_safe=True):
    return IMPL.profile_get_by_name(context, name, project_safe=project_safe)


def profile_get_by_short_id(context, short_id, project_safe=True):
    return IMPL.profile_get_by_short_id(context, short_id,
                                        project_safe=project_safe)


def profile_get_all(context, limit=None, marker=None, sort=None, filters=None,
                    project_safe=True):
    return IMPL.profile_get_all(context, limit=limit, marker=marker,
                                sort=sort, filters=filters,
                                project_safe=project_safe)


def profile_update(context, profile_id, values):
    return IMPL.profile_update(context, profile_id, values)


def profile_delete(context, profile_id):
    return IMPL.profile_delete(context, profile_id)


# Credential
def cred_create(context, values):
    return IMPL.cred_create(context, values)


def cred_get(context, user, project):
    return IMPL.cred_get(context, user, project)


def cred_update(context, user, project, values):
    return IMPL.cred_update(context, user, project, values)


def cred_delete(context, user, project):
    return IMPL.cred_delete(context, user, project)


def cred_create_update(context, values):
    return IMPL.cred_create_update(context, values)


# Events
def event_create(context, values):
    return IMPL.event_create(context, values)


def event_get(context, event_id, project_safe=True):
    return IMPL.event_get(context, event_id, project_safe=project_safe)


def event_get_by_short_id(context, short_id, project_safe=True):
    return IMPL.event_get_by_short_id(context, short_id,
                                      project_safe=project_safe)


def event_get_all(context, limit=None, marker=None, sort=None, filters=None,
                  project_safe=True):
    return IMPL.event_get_all(context, limit=limit, marker=marker, sort=sort,
                              filters=filters, project_safe=project_safe)


def event_count_by_cluster(context, cluster_id, project_safe=True):
    return IMPL.event_count_by_cluster(context, cluster_id,
                                       project_safe=project_safe)


def event_get_all_by_cluster(context, cluster_id, limit=None, marker=None,
                             sort=None, filters=None, project_safe=True):
    return IMPL.event_get_all_by_cluster(context, cluster_id, filters=filters,
                                         limit=limit, marker=marker, sort=sort,
                                         project_safe=project_safe)


def event_prune(context, cluster_id, project_safe=True):
    return IMPL.event_prune(context, cluster_id, project_safe=project_safe)


# Actions
def action_create(context, values):
    return IMPL.action_create(context, values)


def action_update(context, action_id, values):
    return IMPL.action_update(context, action_id, values)


def action_get(context, action_id, project_safe=True, refresh=False):
    return IMPL.action_get(context, action_id, project_safe=project_safe,
                           refresh=refresh)


def action_get_by_name(context, name, project_safe=True):
    return IMPL.action_get_by_name(context, name, project_safe=project_safe)


def action_get_by_short_id(context, short_id, project_safe=True):
    return IMPL.action_get_by_short_id(context, short_id,
                                       project_safe=project_safe)


def action_get_all_by_owner(context, owner):
    return IMPL.action_get_all_by_owner(context, owner)


def action_get_all(context, filters=None, limit=None, marker=None, sort=None,
                   project_safe=True):
    return IMPL.action_get_all(context, filters=filters, sort=sort,
                               limit=limit, marker=marker,
                               project_safe=project_safe)


def action_check_status(context, action_id, timestamp):
    return IMPL.action_check_status(context, action_id, timestamp)


def action_delete_by_target(context, target, action=None,
                            action_excluded=None, status=None):
    return IMPL.action_delete_by_target(context, target, action=action,
                                        action_excluded=action_excluded,
                                        status=status)


def dependency_add(context, depended, dependent):
    return IMPL.dependency_add(context, depended, dependent)


def dependency_get_depended(context, action_id):
    return IMPL.dependency_get_depended(context, action_id)


def dependency_get_dependents(context, action_id):
    return IMPL.dependency_get_dependents(context, action_id)


def action_mark_succeeded(context, action_id, timestamp):
    return IMPL.action_mark_succeeded(context, action_id, timestamp)


def action_mark_ready(context, action_id, timestamp):
    return IMPL.action_mark_ready(context, action_id, timestamp)


def action_mark_failed(context, action_id, timestamp, reason=None):
    return IMPL.action_mark_failed(context, action_id, timestamp, reason)


def action_mark_cancelled(context, action_id, timestamp):
    return IMPL.action_mark_cancelled(context, action_id, timestamp)


def action_acquire(context, action_id, owner, timestamp):
    return IMPL.action_acquire(context, action_id, owner, timestamp)


def action_acquire_random_ready(context, owner, timestamp):
    return IMPL.action_acquire_random_ready(context, owner, timestamp)


def action_acquire_first_ready(context, owner, timestamp):
    return IMPL.action_acquire_first_ready(context, owner, timestamp)


def action_abandon(context, action_id, values=None):
    return IMPL.action_abandon(context, action_id, values)


def action_lock_check(context, action_id, owner=None):
    """Check whether an action has been locked(by an owner)."""
    return IMPL.action_lock_check(context, action_id, owner)


def action_signal(context, action_id, value):
    """Send signal to an action via DB."""
    return IMPL.action_signal(context, action_id, value)


def action_signal_query(context, action_id):
    """Query signal status for the specified action."""
    return IMPL.action_signal_query(context, action_id)


def action_delete(context, action_id):
    return IMPL.action_delete(context, action_id)


def receiver_create(context, values):
    return IMPL.receiver_create(context, values)


def receiver_get(context, receiver_id, project_safe=True):
    return IMPL.receiver_get(context, receiver_id, project_safe=project_safe)


def receiver_get_by_name(context, name, project_safe=True):
    return IMPL.receiver_get_by_name(context, name, project_safe=project_safe)


def receiver_get_by_short_id(context, short_id, project_safe=True):
    return IMPL.receiver_get_by_short_id(context, short_id,
                                         project_safe=project_safe)


def receiver_get_all(context, limit=None, marker=None, filters=None, sort=None,
                     project_safe=True):
    return IMPL.receiver_get_all(context, limit=limit, marker=marker,
                                 sort=sort, filters=filters,
                                 project_safe=project_safe)


def receiver_delete(context, receiver_id):
    return IMPL.receiver_delete(context, receiver_id)


def receiver_update(context, receiver_id, values):
    return IMPL.receiver_update(context, receiver_id, values)


def service_create(service_id, host=None, binary=None, topic=None):
    return IMPL.service_create(service_id, host=host, binary=binary,
                               topic=topic)


def service_update(service_id, values=None):
    return IMPL.service_update(service_id, values=values)


def service_delete(service_id):
    return IMPL.service_delete(service_id)


def service_get(service_id):
    return IMPL.service_get(service_id)


def service_get_all():
    return IMPL.service_get_all()


def gc_by_engine(engine_id):
    return IMPL.gc_by_engine(engine_id)


def registry_create(context, cluster_id, check_type, interval, params,
                    engine_id, enabled=True):
    return IMPL.registry_create(context, cluster_id, check_type, interval,
                                params, engine_id, enabled=enabled)


def registry_update(context, cluster_id, values):
    return IMPL.registry_update(context, cluster_id, values)


def registry_delete(context, cluster_id):
    return IMPL.registry_delete(context, cluster_id)


def registry_claim(context, engine_id):
    return IMPL.registry_claim(context, engine_id)


def registry_get(context, cluster_id):
    return IMPL.registry_get(context, cluster_id)


def registry_get_by_param(context, params):
    return IMPL.registry_get_by_param(context, params)


def db_sync(engine, version=None):
    """Migrate the database to `version` or the most recent version."""
    return IMPL.db_sync(engine, version=version)


def db_version(engine):
    """Display the current database version."""
    return IMPL.db_version(engine)


def event_purge(engine, project, granularity, age):
    """Purge the event records in database."""
    return IMPL.event_purge(project, granularity, age)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from senlin.api.common import version_request as vr
from senlin.api.common import wsgi
from senlin.api.openstack.v1 import version
from senlin.tests.unit.api import shared
from senlin.tests.unit.common import base


class FakeRequest(wsgi.Request):

    @staticmethod
    def blank(*args, **kwargs):
        kwargs['base_url'] = 'http://localhost/v1'
        version = kwargs.pop('version', wsgi.DEFAULT_API_VERSION)
        out = wsgi.Request.blank(*args, **kwargs)
        out.version_request = vr.APIVersionRequest(version)
        return out


class VersionControllerTest(shared.ControllerTest, base.SenlinTestCase):

    def setUp(self):
        super(VersionControllerTest, self).setUp()
        self.controller = version.VersionController({})

    def test_version(self):
        req = self._get('/')

        result = self.controller.version(req)

        response = result['version']
        self.assertEqual('1.0', response['id'])
        self.assertEqual('CURRENT', response['status'])
        self.assertEqual('2016-01-18T00:00:00Z', response['updated'])
        expected = [{
            'base': 'application/json',
            'type': 'application/vnd.openstack.clustering-v1+json'
        }]
        self.assertEqual(expected, response['media-types'])
        expected = [{
            'href': '/v1',
            'rel': 'self'}, {
            'href': 'https://developer.openstack.org/api-ref/clustering',
            'rel': 'help',
        }]
        self.assertEqual(expected, response['links'])


class APIVersionTest(base.SenlinTestCase):

    def setUp(self):
        super(APIVersionTest, self).setUp()
        self.vc = version.VersionController

    def test_min_api_version(self):
        res = self.vc.min_api_version()
        expected = vr.APIVersionRequest(self.vc._MIN_API_VERSION)
        self.assertEqual(expected, res)

    def test_max_api_version(self):
        res = self.vc.max_api_version()
        expected = vr.APIVersionRequest(self.vc._MAX_API_VERSION)
        self.assertEqual(expected, res)

    def test_is_supported(self):
        req = mock.Mock()
        req.version_request = vr.APIVersionRequest(self.vc._MIN_API_VERSION)
        res = self.vc.is_supported(req)
        self.assertTrue(res)

    def test_is_supported_min_version(self):
        req = FakeRequest.blank('/fake', version='1.1')

        self.assertTrue(self.vc.is_supported(req, '1.0', '1.1'))
        self.assertTrue(self.vc.is_supported(req, '1.1', '1.1'))
        self.assertFalse(self.vc.is_supported(req, '1.2'))
        self.assertFalse(self.vc.is_supported(req, '1.3'))

    def test_is_supported_max_version(self):
        req = FakeRequest.blank('/fake', version='2.5')

        self.assertFalse(self.vc.is_supported(req, max_ver='2.4'))
        self.assertTrue(self.vc.is_supported(req, max_ver='2.5'))
        self.assertTrue(self.vc.is_supported(req, max_ver='2.6'))

    def test_is_supported_min_and_max_version(self):
        req = FakeRequest.blank('/fake', version='2.5')

        self.assertFalse(self.vc.is_supported(req, '2.3', '2.4'))
        self.assertTrue(self.vc.is_supported(req, '2.3', '2.5'))
        self.assertTrue(self.vc.is_supported(req, '2.3', '2.7'))
        self.assertTrue(self.vc.is_supported(req, '2.5', '2.7'))
        self.assertFalse(self.vc.is_supported(req, '2.6', '2.7'))
        self.assertTrue(self.vc.is_supported(req, '2.5', '2.5'))
        self.assertFalse(self.vc.is_supported(req, '2.10', '2.1'))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from senlin.db.sqlalchemy import api as db_api
from senlin.db.sqlalchemy import utils as db_utils
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class DBAPIRegistryTest(base.SenlinTestCase):
    def setUp(self):
        super(DBAPIRegistryTest, self).setUp()
        self.ctx = utils.dummy_context()

        db_api.service_create('SERVICE_ID')

    def _create_registry(self, cluster_id, check_type, interval, params,
                         engine_id):
        return db_api.registry_create(self.ctx, cluster_id, check_type,
                                      interval, params, engine_id)

    def test_registry_create_get(self):
        registry = self._create_registry(cluster_id='CLUSTER_ID',
                                         check_type='NODE_STATUS_POLLING',
                                         interval=60,
                                         params={},
                                         engine_id='ENGINE_ID')
        ret_registries = db_api.registry_claim(self.ctx, registry.engine_id)
        self.assertEqual(1, len(ret_registries))
        ret_registry = ret_registries[0]
        self.assertEqual(registry.id, ret_registry.id)
        self.assertEqual(registry.cluster_id, ret_registry.cluster_id)
        self.assertEqual(registry.check_type, ret_registry.check_type)
        self.assertEqual(registry.interval, ret_registry.interval)
        self.assertEqual(registry.params, ret_registry.params)
        self.assertEqual(registry.engine_id, ret_registry.engine_id)

    def test_registry_update(self):
        self._create_registry(cluster_id='FAKE_ID',
                              check_type='NODE_STATUS_POLLING',
                              interval=60,
                              params={},
                              engine_id='DEAD_ENGINE')

        registries = db_api.registry_claim(self.ctx, engine_id='ENGINE_ID')
        self.assertTrue(registries[0].enabled)

        db_api.registry_update(self.ctx, 'FAKE_ID', {'enabled': False})
        registries = db_api.registry_claim(self.ctx, engine_id='NEW_ENGINE_ID')
        self.assertFalse(registries[0].enabled)

    def test_registry_claim(self):
        for i in range(2):
            cluster_id = 'cluster-%s' % i
            self._create_registry(cluster_id=cluster_id,
                                  check_type='NODE_STATUS_POLLING',
                                  interval=60,
                                  params={},
                                  engine_id='DEAD_ENGINE')

        registries = db_api.registry_claim(self.ctx, engine_id='ENGINE_ID')
        self.assertEqual(2, len(registries))
        self.assertEqual('DEAD_ENGINE', registries[0].engine_id)
        self.assertEqual('DEAD_ENGINE', registries[1].engine_id)

    @mock.patch.object(db_utils, 'is_service_dead')
    def test_registry_claim_with_dead_engine(self, mock_check):
        db_api.service_create('SERVICE_ID_DEAD')
        self._create_registry(
            cluster_id='CLUSTER_1', check_type='NODE_STATUS_POLLING',
            interval=60, params={}, engine_id='SERVICE_ID')
        self._create_registry(
            cluster_id='CLUSTER_1', check_type='NODE_STATUS_POLLING',
            interval=60, params={}, engine_id='SERVICE_ID_DEAD')

        mock_check.side_effect = [False, True]

        registries = db_api.registry_claim(self.ctx, engine_id='ENGINE_ID')

        self.assertEqual(1, len(registries))
        self.assertEqual('SERVICE_ID_DEAD', registries[0].engine_id)

    def test_registry_delete(self):
        registry = self._create_registry('CLUSTER_ID',
                                         check_type='NODE_STATUS_POLLING',
                                         interval=60,
                                         params={},
                                         engine_id='ENGINE_ID')
        db_api.registry_delete(self.ctx, 'CLUSTER_ID')
        self.assertEqual([], db_api.registry_claim(self.ctx,
                                                   registry.engine_id))

    def test_registry_get(self):
        obj = self._create_registry(cluster_id='FAKE_ID',
                                    check_type='NODE_STATUS_POLLING',
                                    interval=60,
                                    params={},
                                    engine_id='DEAD_ENGINE')

        registry = db_api.registry_get(self.ctx, 'FAKE_ID')

        self.assertEqual(registry.id, obj.id)
        self.assertEqual(registry.cluster_id, obj.cluster_id)
        self.assertEqual(registry.check_type, obj.check_type)
        self.assertEqual(registry.interval, obj.interval)
        self.assertEqual(registry.params, obj.params)
        self.assertEqual(registry.engine_id, obj.engine_id)

    def test_registry_get_by_engine(self):
        obj = self._create_registry(cluster_id='FAKE_ID',
                                    check_type='NODE_STATUS_POLLING',
                                    interval=60,
                                    params={},
                                    engine_id='ENGINE')

        registry = db_api.registry_get_by_param(
            self.ctx, {"cluster_id": "FAKE_ID", "engine_id": "ENGINE"})

        self.assertEqual(registry.id, obj.id)
        self.assertEqual(registry.cluster_id, obj.cluster_id)
        self.assertEqual(registry.check_type, obj.check_type)
        self.assertEqual(registry.interval, obj.interval)
        self.assertEqual(registry.params, obj.params)
        self.assertEqual(registry.engine_id, obj.engine_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_serialization import jsonutils
from oslo_utils import strutils
from oslo_utils import uuidutils
from oslo_versionedobjects import fields
import re
import six

from senlin.common import consts
from senlin.common.i18n import _

CONF = cfg.CONF

# Field alias for code readability
# BooleanField = fields.BooleanField
FlexibleBooleanField = fields.FlexibleBooleanField
StringField = fields.StringField
IntegerField = fields.IntegerField
FloatField = fields.FloatField
UUIDField = fields.UUIDField
DateTimeField = fields.DateTimeField
DictOfStringsField = fields.DictOfStringsField
ListOfStringsField = fields.ListOfStringsField
ListOfEnumField = fields.ListOfEnumField


class Boolean(fields.FieldType):
    # NOTE: The following definition is much more stricter than the oslo
    #       version. Also note that the treatment of default values here:
    #       we are using the user specified default value when invoking
    #       the 'bool_from_string' until function.

    def __init__(self, default=False):
        super(Boolean, self).__init__()
        self._default = default

    def coerce(self, obj, attr, value):
        return strutils.bool_from_string(value, strict=True,
                                         default=self._default)

    def get_schema(self):
        return {'type': ['boolean']}


class NonNegativeInteger(fields.FieldType):
    # NOTE: This definition is kept because we want the error message from
    #       'int' conversion to be user friendly.
    @staticmethod
    def coerce(obj, attr, value):
        try:
            v = int(value)
        except (TypeError, ValueError):
            raise ValueError(_("The value for %(attr)s must be an integer: "
                               "'%(value)s'.") %
                             {'attr': attr, 'value': value})
        if v < 0:
            err = _("Value must be >= 0 for field '%s'.") % attr
            raise ValueError(err)
        return v

    def get_schema(self):
        return {
            'type': ['integer', 'string'],
            'minimum': 0
        }


# Senlin has a stricter field checking for object fields.
class Object(fields.Object):

    def get_schema(self):
        schema = super(Object, self).get_schema()
        # we are not checking whether self._obj_name is registered, an
        # exception will be raised anyway if it is not registered.
        data_key = 'senlin_object.data'
        schema['properties'][data_key]['additionalProperties'] = False
        return schema


class UUID(fields.FieldType):

    _PATTERN = (r'^[a-fA-F0-9]{8}-?[a-fA-F0-9]{4}-?[a-fA-F0-9]{4}-?[a-fA-F0-9]'
                r'{4}-?[a-fA-F0-9]{12}$')

    @staticmethod
    def coerce(obj, attr, value):
        if not uuidutils.is_uuid_like(value):
            msg = _("The value for %(attr)s is not a valid UUID: '%(value)s'."
                    ) % {'attr': attr, 'value': value}
            raise ValueError(msg)

        return str(value)

    def get_schema(self):
        return {'type': ['string'], 'pattern': self._PATTERN}


class Json(fields.FieldType):
    def coerce(self, obj, attr, value):
        if isinstance(value, six.string_types):
            try:
                return jsonutils.loads(value)
            except ValueError:
                msg = _("The value (%s) is not a valid JSON.") % value
                raise ValueError(msg)
        return value

    def from_primitive(self, obj, attr, value):
        return self.coerce(obj, attr, value)

    def to_primitive(self, obj, attr, value):
        return jsonutils.dumps(value)

    def stringify(self, value):
        if isinstance(value, six.string_types):
            try:
                return jsonutils.loads(value)
            except ValueError:
                raise
        return str(value)

    def get_schema(self):
        return {'type': ['object']}


class NotificationPriority(fields.Enum):

    # The priorities here are derived from oslo_messaging.notify.notifier
    ALL = consts.NOTIFICATION_PRIORITIES

    def __init__(self):
        super(NotificationPriority, self).__init__(self.ALL)


class NotificationPhase(fields.Enum):

    ALL = consts.NOTIFICATION_PHASES

    def __init__(self):
        super(NotificationPhase, self).__init__(self.ALL)


class Name(fields.String):

    def __init__(self, min_len=1, max_len=255):
        super(Name, self).__init__()
        self.min_len = min_len
        self.max_len = max_len

    def coerce(self, obj, attr, value):
        err = None
        if len(value) < self.min_len:
            err = _("The value for the %(attr)s field must be at least "
                    "%(count)d characters long."
                    ) % {'attr': attr, 'count': self.min_len}
        elif len(value) > self.max_len:
            err = _("The value for the %(attr)s field must be less than "
                    "%(count)d characters long."
                    ) % {'attr': attr, 'count': self.max_len}
        else:
            # NOTE: This is pretty restrictive. We can relax it later when
            # there are requests to do so
            regex = re.compile(u'^[a-zA-Z\u4e00-\u9fa5\d\.\_\~-]*$',
                               re.IGNORECASE)
            if not regex.search(value):
                err = _("The value for the '%(attr)s' (%(value)s) contains "
                        "illegal characters. It must contain only "
                        "alphanumeric or \"_-.~\" characters and must start "
                        "with letter."
                        ) % {'attr': attr, 'value': value}

        if err:
            raise ValueError(err)

        return super(Name, self).coerce(obj, attr, value)

    def get_schema(self):
        return {
            'type': ['string'],
            'minLength': self.min_len,
            'maxLength': self.max_len
        }


class Capacity(fields.Integer):

    def __init__(self, minimum=0, maximum=None):
        super(Capacity, self).__init__()
        CONF.import_opt("max_nodes_per_cluster", "senlin.common.config")

        if minimum > CONF.max_nodes_per_cluster:
            err = _("The value of 'minimum' cannot be greater than the global "
                    "constraint (%(m)d).") % {'m': CONF.max_nodes_per_cluster}
            raise ValueError(err)
        self.minimum = minimum

        if maximum is not None:
            if maximum < minimum:
                err = _("The value of 'maximum' must be greater than or equal "
                        "to that of the 'minimum' specified.")
                raise ValueError(err)

            if maximum > CONF.max_nodes_per_cluster:
                err = _("The value of 'maximum' cannot be greater than the "
                        "global constraint (%(m)d)."
                        ) % {'m': CONF.max_nodes_per_cluster}
                raise ValueError(err)

            self.maximum = maximum
        else:
            self.maximum = CONF.max_nodes_per_cluster

    def coerce(self, obj, attr, value):
        try:
            v = int(value)
        except Exception:
            raise ValueError(_("The value for %(attr)s must be an integer: "
                               "'%(value)s'.") %
                             {'attr': attr, 'value': value})
        if v < self.minimum:
            raise ValueError(_("The value for the %(a)s field must be greater "
                               "than or equal to %(n)d.") %
                             {'a': attr, 'n': self.minimum})
        elif v > self.maximum:
            raise ValueError(_("The value for the %(a)s field must be less "
                               "than or equal to %(n)d.") %
                             {'a': attr, 'n': self.maximum})
        return super(Capacity, self).coerce(obj, attr, v)

    def get_schema(self):
        return {
            'type': ['integer', 'string'],
            'minimum': self.minimum,
            'maximum': self.maximum,
            'pattern': '^[0-9]*$',
        }


class Sort(fields.String):

    def __init__(self, valid_keys):
        super(Sort, self).__init__()
        self.valid_keys = valid_keys

    def coerce(self, obj, attr, value):
        for s in value.split(','):
            s_key, _sep, s_dir = s.partition(':')
            err = None
            if not s_key:
                err = _("Missing sort key for '%s'.") % attr
                raise ValueError(err)

            if s_key not in self.valid_keys:
                err = _("Unsupported sort key '%(value)s' for '%(attr)s'."
                        ) % {'attr': attr, 'value': s_key}

            if s_dir and s_dir not in ('asc', 'desc'):
                err = _("Unsupported sort dir '%(value)s' for '%(attr)s'."
                        ) % {'attr': attr, 'value': s_dir}

            if err:
                raise ValueError(err)

        return super(Sort, self).coerce(obj, attr, value)

    def get_schema(self):
        return {
            'type': ['string'],
        }


class IdentityList(fields.List):

    def __init__(self, element_type, min_items=0, unique=True, nullable=False,
                 **kwargs):
        super(IdentityList, self).__init__(element_type, **kwargs)
        self.min_items = min_items
        self.unique_items = unique
        self.nullable = nullable

    def coerce(self, obj, attr, value):
        res = super(IdentityList, self).coerce(obj, attr, value)
        if len(res) < self.min_items:
            raise ValueError(_("Value for '%(attr)s' must have at least "
                               "%(num)s item(s).") %
                             {'attr': attr, 'num': self.min_items})
        if len(set(res)) != len(res) and self.unique_items:
            raise ValueError(_("Items for '%(attr)s' must be unique") %
                             {'attr': attr})
        return res

    def get_schema(self):
        schema = super(IdentityList, self).get_schema()
        if self.nullable:
            schema['type'].append('null')
        schema['minItems'] = self.min_items
        schema['uniqueItems'] = self.unique_items
        return schema


class BaseEnum(fields.FieldType):
    # NOTE: We are not basing Enum on String because String is not working
    # correctly when handling None value.
    def __init__(self, nullable=False):
        valid_values = list(self.__class__.ALL)
        if not valid_values:
            raise ValueError(_("No list of valid values provided for enum."))

        for value in valid_values:
            if not isinstance(value, six.string_types):
                raise ValueError(_("Enum field only support string values."))

        self._valid_values = list(valid_values)
        self._nullable = nullable
        super(BaseEnum, self).__init__()

    def coerce(self, obj, attr, value):
        value = six.text_type(value)
        if value not in self._valid_values:
            raise ValueError(_("Value '%(value)s' is not acceptable for "
                               "field '%(attr)s'.") %
                             {'value': value, 'attr': attr})
        return value

    def stringify(self, value):
        if value is None:
            return None
        return '\'%s\'' % value


class AdjustmentType(BaseEnum):

    ALL = consts.ADJUSTMENT_TYPES

    def get_schema(self):
        return {'type': ['string'],
                'enum': self._valid_values}


class ClusterActionName(BaseEnum):

    ALL = consts.CLUSTER_ACTION_NAMES

    def get_schema(self):
        return {'type': ['string'],
                'enum': self._valid_values}


class ClusterStatus(BaseEnum):

    ALL = consts.CLUSTER_STATUSES


class NodeStatus(BaseEnum):

    ALL = consts.NODE_STATUSES


class ActionStatus(BaseEnum):

    ALL = consts.ACTION_STATUSES


class ReceiverType(BaseEnum):

    ALL = consts.RECEIVER_TYPES

    def get_schema(self):
        return {'type': ['string'],
                'enum': self._valid_values}


class UniqueDict(fields.Dict):

    def coerce(self, obj, attr, value):
        res = super(UniqueDict, self).coerce(obj, attr, value)
        new_nodes = res.values()
        if len(new_nodes) != len(set(new_nodes)):
            raise ValueError(_("Map contains duplicated values"))
        return res


# TODO(Qiming): remove this when oslo patch is released
# https://review.openstack.org/#/c/360095
class NonNegativeIntegerField(fields.AutoTypedField):

    AUTO_TYPE = NonNegativeInteger()


class BooleanField(fields.AutoTypedField):

    AUTO_TYPE = Boolean()


# An override to the oslo.versionedobjects version so that we are using
# our own Object definition.
class ObjectField(fields.AutoTypedField):

    def __init__(self, objtype, subclasses=False, **kwargs):
        self.AUTO_TYPE = Object(objtype, subclasses)
        self.objname = objtype
        super(ObjectField, self).__init__(**kwargs)


class JsonField(fields.AutoTypedField):
    AUTO_TYPE = Json()


class ListField(fields.AutoTypedField):
    AUTO_TYPE = fields.List(fields.FieldType())


class NotificationPriorityField(fields.BaseEnumField):
    AUTO_TYPE = NotificationPriority()


class NotificationPhaseField(fields.BaseEnumField):
    AUTO_TYPE = NotificationPhase()


class NameField(fields.AutoTypedField):

    AUTO_TYPE = Name()


class UUIDField(fields.AutoTypedField):

    AUTO_TYPE = UUID()


class CapacityField(fields.AutoTypedField):

    AUTO_TYPE = None

    def __init__(self, nullable=False, default=None, minimum=0, maximum=None):
        self.AUTO_TYPE = Capacity(minimum=minimum, maximum=maximum)
        super(CapacityField, self).__init__(nullable=nullable, default=default)


class SortField(fields.AutoTypedField):

    AUTO_TYPE = None

    def __init__(self, valid_keys, nullable=False, default=None):
        self.AUTO_TYPE = Sort(valid_keys)
        super(SortField, self).__init__(nullable=nullable, default=default)


class IdentityListField(fields.AutoTypedField):

    AUTO_TYPE = None

    def __init__(self, min_items=0, unique=True, nullable=False, default=None):
        if default is None:
            default = []
        self.AUTO_TYPE = IdentityList(fields.String(), min_items=min_items,
                                      unique=unique)
        super(IdentityListField, self).__init__(nullable=nullable,
                                                default=default)


class AdjustmentTypeField(fields.AutoTypedField):

    AUTO_TYPE = None

    def __init__(self, **kwargs):
        nullable = kwargs.get('nullable', False)
        self.AUTO_TYPE = AdjustmentType(nullable=nullable)
        super(AdjustmentTypeField, self).__init__(**kwargs)


class ClusterActionNameField(fields.AutoTypedField):

    AUTO_TYPE = None

    def __init__(self, **kwargs):
        nullable = kwargs.get('nullable', False)
        self.AUTO_TYPE = ClusterActionName(nullable=nullable)
        super(ClusterActionNameField, self).__init__(**kwargs)


class ClusterStatusField(fields.AutoTypedField):

    AUTO_TYPE = ClusterStatus


class NodeStatusField(fields.AutoTypedField):

    AUTO_TYPE = NodeStatus


class ActionStatusField(fields.AutoTypedField):

    AUTO_TYPE = ActionStatus


class ReceiverTypeField(fields.AutoTypedField):

    AUTO_TYPE = None

    def __init__(self, **kwargs):
        nullable = kwargs.get('nullable', False)
        self.AUTO_TYPE = ReceiverType(nullable=nullable)
        super(ReceiverTypeField, self).__init__(**kwargs)


class NodeReplaceMapField(fields.AutoTypedField):

    AUTO_TYPE = UniqueDict(fields.String())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from senlin.drivers.os import octavia_v2
from senlin.drivers import sdk
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class TestOctaviaV2Driver(base.SenlinTestCase):

    def setUp(self):
        super(TestOctaviaV2Driver, self).setUp()
        self.context = utils.dummy_context()
        self.conn_params = self.context.to_dict()
        self.conn = mock.Mock()
        with mock.patch.object(sdk, 'create_connection') as mock_creare_conn:
            mock_creare_conn.return_value = self.conn
            self.oc = octavia_v2.OctaviaClient(self.context)

    @mock.patch.object(sdk, 'create_connection')
    def test_init(self, mock_create_connection):
        params = self.conn_params
        octavia_v2.OctaviaClient(params)
        mock_create_connection.assert_called_once_with(params)

    def test_loadbalancer_get(self):
        lb_id = 'loadbalancer_identifier'
        loadbalancer_obj = mock.Mock()

        self.conn.load_balancer.find_load_balancer.return_value = \
            loadbalancer_obj
        res = self.oc.loadbalancer_get(lb_id)
        self.conn.load_balancer.find_load_balancer.assert_called_once_with(
            lb_id, False)
        self.assertEqual(loadbalancer_obj, res)

    def test_loadbalancer_create(self):
        vip_subnet_id = 'ID1'
        lb_obj = mock.Mock()

        # All input parameters are provided
        kwargs = {
            'vip_address': '192.168.0.100',
            'name': 'test-loadbalancer',
            'description': 'This is a loadbalancer',
            'admin_state_up': True
        }

        self.conn.load_balancer.create_load_balancer.return_value = lb_obj
        self.assertEqual(lb_obj, self.oc.loadbalancer_create(vip_subnet_id,
                                                             **kwargs))
        self.conn.load_balancer.create_load_balancer.assert_called_once_with(
            vip_subnet_id=vip_subnet_id, **kwargs)

        # Use default input parameters
        kwargs = {
            'admin_state_up': True
        }
        self.assertEqual(lb_obj, self.oc.loadbalancer_create(vip_subnet_id))
        self.conn.load_balancer.create_load_balancer.assert_called_with(
            vip_subnet_id=vip_subnet_id, **kwargs)

    def test_loadbalancer_delete(self):
        lb_id = 'ID1'

        self.oc.loadbalancer_delete(lb_id, ignore_missing=False)
        self.conn.load_balancer.delete_load_balancer.assert_called_once_with(
            lb_id, ignore_missing=False)

        self.oc.loadbalancer_delete(lb_id)
        self.conn.load_balancer.delete_load_balancer.assert_called_with(
            lb_id, ignore_missing=True)

    def test_listener_create(self):
        loadbalancer_id = 'ID1'
        protocol = 'HTTP'
        protocol_port = 80
        listener_obj = mock.Mock()

        # All input parameters are provided
        kwargs = {
            'connection_limit': 100,
            'admin_state_up': True,
            'name': 'test-listener',
            'description': 'This is a listener',
        }

        self.conn.load_balancer.create_listener.return_value = listener_obj
        self.assertEqual(listener_obj, self.oc.listener_create(
            loadbalancer_id, protocol, protocol_port, **kwargs))
        self.conn.load_balancer.create_listener.assert_called_once_with(
            loadbalancer_id=loadbalancer_id, protocol=protocol,
            protocol_port=protocol_port, **kwargs)

        # Use default input parameters
        kwargs = {
            'admin_state_up': True
        }
        self.assertEqual(listener_obj, self.oc.listener_create(
            loadbalancer_id, protocol, protocol_port))
        self.conn.load_balancer.create_listener.assert_called_with(
            loadbalancer_id=loadbalancer_id, protocol=protocol,
            protocol_port=protocol_port, **kwargs)

    def test_listener_delete(self):
        listener_id = 'ID1'

        self.oc.listener_delete(listener_id, ignore_missing=False)
        self.conn.load_balancer.delete_listener.assert_called_once_with(
            listener_id, ignore_missing=False)

        self.oc.listener_delete(listener_id)
        self.conn.load_balancer.delete_listener.assert_called_with(
            listener_id, ignore_missing=True)

    def test_pool_create(self):
        lb_algorithm = 'ROUND_ROBIN'
        listener_id = 'ID1'
        protocol = 'HTTP'
        pool_obj = mock.Mock()

        # All input parameters are provided
        kwargs = {
            'admin_state_up': True,
            'name': 'test-pool',
            'description': 'This is a pool',
        }

        self.conn.load_balancer.create_pool.return_value = pool_obj
        self.assertEqual(pool_obj, self.oc.pool_create(
            lb_algorithm, listener_id, protocol, **kwargs))
        self.conn.load_balancer.create_pool.assert_called_once_with(
            lb_algorithm=lb_algorithm, listener_id=listener_id,
            protocol=protocol, **kwargs)

        # Use default input parameters
        kwargs = {
            'admin_state_up': True
        }
        self.assertEqual(pool_obj, self.oc.pool_create(
            lb_algorithm, listener_id, protocol))
        self.conn.load_balancer.create_pool.assert_called_with(
            lb_algorithm=lb_algorithm, listener_id=listener_id,
            protocol=protocol, **kwargs)

    def test_pool_delete(self):
        pool_id = 'ID1'

        self.oc.pool_delete(pool_id, ignore_missing=False)
        self.conn.load_balancer.delete_pool.assert_called_once_with(
            pool_id, ignore_missing=False)

        self.oc.pool_delete(pool_id)
        self.conn.load_balancer.delete_pool.assert_called_with(
            pool_id, ignore_missing=True)

    def test_pool_member_create(self):
        pool_id = 'ID1'
        address = '192.168.1.100'
        protocol_port = 80
        subnet_id = 'ID2'
        weight = 50
        member_obj = mock.Mock()

        # All input parameters are provided
        kwargs = {
            'weight': weight,
            'admin_state_up': True,
        }

        self.conn.load_balancer.create_member.return_value = member_obj
        self.assertEqual(member_obj, self.oc.pool_member_create(
            pool_id, address, protocol_port, subnet_id, **kwargs))
        self.conn.load_balancer.create_member.assert_called_once_with(
            pool_id, address=address, protocol_port=protocol_port,
            subnet_id=subnet_id, **kwargs)

        # Use default input parameters
        kwargs = {
            'admin_state_up': True
        }
        self.assertEqual(member_obj, self.oc.pool_member_create(
            pool_id, address, protocol_port, subnet_id))
        self.conn.load_balancer.create_member.assert_called_with(
            pool_id, address=address, protocol_port=protocol_port,
            subnet_id=subnet_id, **kwargs)

    def test_pool_member_delete(self):
        pool_id = 'ID1'
        member_id = 'ID2'

        self.oc.pool_member_delete(pool_id, member_id, ignore_missing=False)
        self.conn.load_balancer.delete_member.assert_called_once_with(
            member_id, pool_id, ignore_missing=False)

        self.oc.pool_member_delete(pool_id, member_id)
        self.conn.load_balancer.delete_member.assert_called_with(
            member_id, pool_id, ignore_missing=True)

    def test_healthmonitor_create(self):
        hm_type = 'HTTP'
        delay = 30
        timeout = 10
        max_retries = 5
        pool_id = 'ID1'
        hm_obj = mock.Mock()

        # All input parameters are provided
        kwargs = {
            'http_method': 'test-method',
            'admin_state_up': True,
            'url_path': '/test_page',
            'expected_codes': [200, 201, 202],
        }

        self.conn.load_balancer.create_health_monitor.return_value = hm_obj
        res = self.oc.healthmonitor_create(hm_type, delay, timeout,
                                           max_retries, pool_id, **kwargs)
        self.assertEqual(hm_obj, res)
        self.conn.load_balancer.create_health_monitor.assert_called_once_with(
            type=hm_type, delay=delay, timeout=timeout,
            max_retries=max_retries, pool_id=pool_id, **kwargs)

        # Use default input parameters
        res = self.oc.healthmonitor_create(hm_type, delay, timeout,
                                           max_retries, pool_id,
                                           admin_state_up=True)
        self.assertEqual(hm_obj, res)
        self.conn.load_balancer.create_health_monitor.assert_called_with(
            type=hm_type, delay=delay, timeout=timeout,
            max_retries=max_retries, pool_id=pool_id,
            admin_state_up=True)

        # hm_type other than HTTP, then other params ignored
        res = self.oc.healthmonitor_create('TCP', delay, timeout,
                                           max_retries, pool_id, **kwargs)
        self.assertEqual(hm_obj, res)
        self.conn.load_balancer.create_health_monitor.assert_called_with(
            type='TCP', delay=delay, timeout=timeout,
            max_retries=max_retries, pool_id=pool_id,
            admin_state_up=True)

    def test_healthmonitor_delete(self):
        healthmonitor_id = 'ID1'

        self.oc.healthmonitor_delete(healthmonitor_id, ignore_missing=False)
        self.conn.load_balancer.delete_health_monitor.assert_called_once_with(
            healthmonitor_id, ignore_missing=False)

        self.oc.healthmonitor_delete(healthmonitor_id)
        self.conn.load_balancer.delete_health_monitor.assert_called_with(
            healthmonitor_id, ignore_missing=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import six

from senlin.common import exception as exc
from senlin.profiles.os.nova import server
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils

spec = {
    'type': 'os.nova.server',
    'version': '1.0',
    'properties': {
        'context': {},
        'auto_disk_config': True,
        'availability_zone': 'FAKE_AZ',
        'block_device_mapping': [{
            'device_name': 'FAKE_NAME',
            'volume_size': 1000,
        }],
        'flavor': 'FLAV',
        'image': 'FAKE_IMAGE',
        'key_name': 'FAKE_KEYNAME',
        "metadata": {"meta var": "meta val"},
        'name': 'FAKE_SERVER_NAME',
        'networks': [{
            'floating_ip': 'FAKE_FLOATING_IP',
            'floating_network': 'FAKE_FLOATING_NET',
            'security_groups': ['FAKE_SECURITY_GROUP'],
            'port': 'FAKE_PORT',
            'fixed_ip': 'FAKE_IP',
            'network': 'FAKE_NET',
        }],
        'scheduler_hints': {
            'same_host': 'HOST_ID',
        },
    }
}


class TestAvailabilityZoneValidation(base.SenlinTestCase):

    scenarios = [
        ('validate:success', dict(
            reason=None,
            success=True,
            validate_result=[['FAKE_AZ']],
            result='FAKE_AZ',
            exception=None,
            message='')),
        ('validate:driver_failure', dict(
            reason=None,
            success=False,
            validate_result=exc.InternalError(message='BANG.'),
            result='FAKE_AZ',
            exception=exc.InternalError,
            message='BANG.')),
        ('validate:not_found', dict(
            reason=None,
            success=False,
            validate_result=[[]],
            result='FAKE_AZ',
            exception=exc.InvalidSpec,
            message=("The specified availability_zone 'FAKE_AZ' could "
                     "not be found"))),
        ('create:success', dict(
            reason='create',
            success=True,
            validate_result=[['FAKE_AZ']],
            result='FAKE_AZ',
            exception=None,
            message='')),
        ('create:driver_failure', dict(
            reason='create',
            success=False,
            validate_result=exc.InternalError(message='BANG'),
            result='FAKE_AZ',
            exception=exc.EResourceCreation,
            message='Failed in creating server: BANG.')),
        ('create:not_found', dict(
            reason='create',
            success=False,
            validate_result=[[]],
            result='FAKE_AZ',
            exception=exc.EResourceCreation,
            message=("Failed in creating server: The specified "
                     "availability_zone 'FAKE_AZ' could not be found.")))
    ]

    def setUp(self):
        super(TestAvailabilityZoneValidation, self).setUp()

        self.cc = mock.Mock()
        prof = server.ServerProfile('t', spec)
        prof._computeclient = self.cc
        self.profile = prof

    def test_validation(self):
        self.cc.validate_azs.side_effect = self.validate_result
        node = mock.Mock(id='NODE_ID')

        if self.success:
            res = self.profile._validate_az(node, 'FAKE_AZ', self.reason)
            self.assertEqual(self.result, res)
        else:
            ex = self.assertRaises(self.exception,
                                   self.profile._validate_az,
                                   node, 'FAKE_AZ', self.reason)
            self.assertEqual(self.message, six.text_type(ex))

        self.cc.validate_azs.assert_called_once_with(['FAKE_AZ'])


class TestFlavorValidation(base.SenlinTestCase):

    scenarios = [
        ('validate:success', dict(
            reason=None,
            success=True,
            validate_result=[mock.Mock(id='FID', is_disabled=False)],
            result='FID',
            exception=None,
            message='')),
        ('validate:driver_failure', dict(
            reason=None,
            success=False,
            validate_result=exc.InternalError(message='BANG.'),
            result='FID',
            exception=exc.InternalError,
            message='BANG.')),
        ('validate:not_found', dict(
            reason=None,
            success=False,
            validate_result=exc.InternalError(code=404, message='BANG.'),
            result='FID',
            exception=exc.InvalidSpec,
            message="The specified flavor 'FLAVOR' could not be found.")),
        ('validate:disabled', dict(
            reason=None,
            success=False,
            validate_result=[mock.Mock(id='FID', is_disabled=True)],
            result='FID',
            exception=exc.InvalidSpec,
            message="The specified flavor 'FLAVOR' is disabled")),
        ('create:success', dict(
            reason='create',
            success=True,
            validate_result=[mock.Mock(id='FID', is_disabled=False)],
            result='FID',
            exception=None,
            message='')),
        ('create:driver_failure', dict(
            reason='create',
            success=False,
            validate_result=exc.InternalError(message='BANG'),
            result='FID',
            exception=exc.EResourceCreation,
            message='Failed in creating server: BANG.')),
        ('create:not_found', dict(
            reason='create',
            success=False,
            validate_result=exc.InternalError(code=404, message='BANG'),
            result='FID',
            exception=exc.EResourceCreation,
            message="Failed in creating server: BANG.")),
        ('create:disabled', dict(
            reason='create',
            success=False,
            validate_result=[mock.Mock(id='FID', is_disabled=True)],
            result='FID',
            exception=exc.EResourceCreation,
            message=("Failed in creating server: The specified flavor "
                     "'FLAVOR' is disabled."))),
        ('update:success', dict(
            reason='update',
            success=True,
            validate_result=[mock.Mock(id='FID', is_disabled=False)],
            result='FID',
            exception=None,
            message='')),
        ('update:driver_failure', dict(
            reason='update',
            success=False,
            validate_result=exc.InternalError(message='BANG'),
            result='FID',
            exception=exc.EResourceUpdate,
            message="Failed in updating server 'NOVA_ID': BANG.")),
        ('update:not_found', dict(
            reason='update',
            success=False,
            validate_result=exc.InternalError(code=404, message='BANG'),
            result='FID',
            exception=exc.EResourceUpdate,
            message="Failed in updating server 'NOVA_ID': BANG.")),
        ('update:disabled', dict(
            reason='update',
            success=False,
            validate_result=[mock.Mock(id='FID', is_disabled=True)],
            result='FID',
            exception=exc.EResourceUpdate,
            message=("Failed in updating server 'NOVA_ID': The specified "
                     "flavor 'FLAVOR' is disabled.")))
    ]

    def setUp(self):
        super(TestFlavorValidation, self).setUp()

        self.cc = mock.Mock()
        self.profile = server.ServerProfile('t', spec)
        self.profile._computeclient = self.cc

    def test_validation(self):
        self.cc.flavor_find.side_effect = self.validate_result
        node = mock.Mock(id='NODE_ID', physical_id='NOVA_ID')
        flavor = 'FLAVOR'

        if self.success:
            res = self.profile._validate_flavor(node, flavor, self.reason)
            self.assertIsNotNone(res)
            self.assertEqual(self.result, res.id)
        else:
            ex = self.assertRaises(self.exception,
                                   self.profile._validate_flavor,
                                   node, flavor, self.reason)
            self.assertEqual(self.message, six.text_type(ex))

        self.cc.flavor_find.assert_called_once_with(flavor, False)


class TestImageValidation(base.SenlinTestCase):

    scenarios = [
        ('validate:success', dict(
            reason=None,
            success=True,
            validate_result=[mock.Mock(id='IMAGE_ID')],
            result='IMAGE_ID',
            exception=None,
            message='')),
        ('validate:driver_failure', dict(
            reason=None,
            success=False,
            validate_result=exc.InternalError(message='BANG.'),
            result='FID',
            exception=exc.InternalError,
            message='BANG.')),
        ('validate:not_found', dict(
            reason=None,
            success=False,
            validate_result=exc.InternalError(code=404, message='BANG.'),
            result='FID',
            exception=exc.InvalidSpec,
            message="The specified image 'IMAGE' could not be found.")),
        ('create:success', dict(
            reason='create',
            success=True,
            validate_result=[mock.Mock(id='IMAGE_ID')],
            result='IMAGE_ID',
            exception=None,
            message='')),
        ('create:driver_failure', dict(
            reason='create',
            success=False,
            validate_result=exc.InternalError(message='BANG'),
            result='FID',
            exception=exc.EResourceCreation,
            message='Failed in creating server: BANG.')),
        ('create:not_found', dict(
            reason='create',
            success=False,
            validate_result=exc.InternalError(code=404, message='BANG'),
            result='FID',
            exception=exc.EResourceCreation,
            message="Failed in creating server: BANG.")),
        ('update:success', dict(
            reason='update',
            success=True,
            validate_result=[mock.Mock(id='IMAGE_ID')],
            result='IMAGE_ID',
            exception=None,
            message='')),
        ('update:driver_failure', dict(
            reason='update',
            success=False,
            validate_result=exc.InternalError(message='BANG'),
            result='FID',
            exception=exc.EResourceUpdate,
            message="Failed in updating server 'NOVA_ID': BANG.")),
        ('update:not_found', dict(
            reason='update',
            success=False,
            validate_result=exc.InternalError(code=404, message='BANG'),
            result='FID',
            exception=exc.EResourceUpdate,
            message="Failed in updating server 'NOVA_ID': BANG.")),
    ]

    def setUp(self):
        super(TestImageValidation, self).setUp()

        self.cc = mock.Mock()
        self.profile = server.ServerProfile('t', spec)
        self.profile._computeclient = self.cc

    def test_validation(self):
        self.cc.image_find.side_effect = self.validate_result
        node = mock.Mock(id='NODE_ID', physical_id='NOVA_ID')
        image = 'IMAGE'

        if self.success:
            res = self.profile._validate_image(node, image, self.reason)
            self.assertIsNotNone(res)
            self.assertEqual(self.result, res.id)
        else:
            ex = self.assertRaises(self.exception,
                                   self.profile._validate_image,
                                   node, image, self.reason)
            self.assertEqual(self.message, six.text_type(ex))

        self.cc.image_find.assert_called_once_with(image, False)


class TestVolumeValidation(base.SenlinTestCase):

    scenarios = [
        ('validate:success', dict(
            reason=None,
            success=True,
            validate_result=[mock.Mock(id='VOLUME_ID', status='available')],
            result='VOLUME_ID',
            exception=None,
            message='')),
        ('validate:failure', dict(
            reason=None,
            success=False,
            validate_result=[mock.Mock(id='VOLUME_ID', status='in-use')],
            result='VOLUME_ID',
            exception=exc.InvalidSpec,
            message="The volume VOLUME should be in 'available' "
                    "status but is in 'in-use' status.")),
        ('validate:driver_failure', dict(
            reason=None,
            success=False,
            validate_result=exc.InternalError(message='BANG.'),
            result='FID',
            exception=exc.InternalError,
            message='BANG.')),
        ('validate:not_found', dict(
            reason=None,
            success=False,
            validate_result=exc.InternalError(code=404, message='BANG.'),
            result='FID',
            exception=exc.InvalidSpec,
            message="The specified volume 'VOLUME' could not be found.")),
        ('create:success', dict(
            reason='create',
            success=True,
            validate_result=[mock.Mock(id='VOLUME_ID', status='available')],
            result='VOLUME_ID',
            exception=None,
            message='')),
        ('create:driver_failure', dict(
            reason='create',
            success=False,
            validate_result=exc.InternalError(message='BANG'),
            result='FID',
            exception=exc.EResourceCreation,
            message='Failed in creating server: BANG.')),
        ('create:not_found', dict(
            reason='create',
            success=False,
            validate_result=exc.InternalError(code=404, message='BANG'),
            result='FID',
            exception=exc.EResourceCreation,
            message="Failed in creating server: BANG.")),
    ]

    def setUp(self):
        super(TestVolumeValidation, self).setUp()
        bdm_v2 = [
            {
                'volume_size': 1,
                'uuid': '6ce0be68',
                'source_type': 'volume',
                'destination_type': 'volume',
                'boot_index': 0,
            },
        ]

        volume_spec = {
            'type': 'os.nova.server',
            'version': '1.0',
            'properties': {
                'flavor': 'FLAV',
                'name': 'FAKE_SERVER_NAME',
                'security_groups': ['HIGH_SECURITY_GROUP'],
                'block_device_mapping_v2': bdm_v2,
            }
        }

        self.vc = mock.Mock()
        self.profile = server.ServerProfile('t', volume_spec)
        self.profile._block_storageclient = self.vc

    def test_validation(self):
        self.vc.volume_get.side_effect = self.validate_result
        node = mock.Mock(id='NODE_ID', physical_id='NOVA_ID')
        volume = 'VOLUME'

        if self.success:
            res = self.profile._validate_volume(node, volume, self.reason)
            self.assertIsNotNone(res)
            self.assertEqual(self.result, res.id)
        else:
            ex = self.assertRaises(self.exception,
                                   self.profile._validate_volume,
                                   node, volume, self.reason)
            self.assertEqual(self.message, six.text_type(ex))

        self.vc.volume_get.assert_called_once_with(volume)


class TestKeypairValidation(base.SenlinTestCase):

    scenarios = [
        ('validate:success', dict(
            reason=None,
            success=True,
            validate_result=[mock.Mock(id='KEY_ID')],
            result='KEY_ID',
            exception=None,
            message='')),
        ('validate:driver_failure', dict(
            reason=None,
            success=False,
            validate_result=exc.InternalError(message='BANG.'),
            result='FID',
            exception=exc.InternalError,
            message='BANG.')),
        ('validate:not_found', dict(
            reason=None,
            success=False,
            validate_result=exc.InternalError(code=404, message='BANG.'),
            result='FID',
            exception=exc.InvalidSpec,
            message="The specified key_name 'KEY' could not be found.")),
        ('create:success', dict(
            reason='create',
            success=True,
            validate_result=[mock.Mock(id='IMAGE_ID')],
            result='IMAGE_ID',
            exception=None,
            message='')),
        ('create:driver_failure', dict(
            reason='create',
            success=False,
            validate_result=exc.InternalError(message='BANG'),
            result='FID',
            exception=exc.EResourceCreation,
            message='Failed in creating server: BANG.')),
        ('create:not_found', dict(
            reason='create',
            success=False,
            validate_result=exc.InternalError(code=404, message='BANG'),
            result='FID',
            exception=exc.EResourceCreation,
            message="Failed in creating server: BANG.")),
        ('update:success', dict(
            reason='update',
            success=True,
            validate_result=[mock.Mock(id='KEY_ID')],
            result='KEY_ID',
            exception=None,
            message='')),
        ('update:driver_failure', dict(
            reason='update',
            success=False,
            validate_result=exc.InternalError(message='BANG'),
            result='FID',
            exception=exc.EResourceUpdate,
            message="Failed in updating server 'NOVA_ID': BANG.")),
        ('update:not_found', dict(
            reason='update',
            success=False,
            validate_result=exc.InternalError(code=404, message='BANG'),
            result='FID',
            exception=exc.EResourceUpdate,
            message="Failed in updating server 'NOVA_ID': BANG.")),
    ]

    def setUp(self):
        super(TestKeypairValidation, self).setUp()

        self.cc = mock.Mock()
        self.profile = server.ServerProfile('t', spec)
        self.profile._computeclient = self.cc

    def test_validation(self):
        self.cc.keypair_find.side_effect = self.validate_result
        node = mock.Mock(id='NODE_ID', physical_id='NOVA_ID')
        key = 'KEY'

        if self.success:
            res = self.profile._validate_keypair(node, key, self.reason)
            self.assertIsNotNone(res)
            self.assertEqual(self.result, res.id)
        else:
            ex = self.assertRaises(self.exception,
                                   self.profile._validate_keypair,
                                   node, key, self.reason)
            self.assertEqual(self.message, six.text_type(ex))

        self.cc.keypair_find.assert_called_once_with(key, False)


class TestNetworkValidation(base.SenlinTestCase):

    scenarios = [
        ('validate:net-n:port-n:fixed_ip-n:sgroups-n', dict(
            reason=None,
            success=True,
            inputs={'port': 'PORT'},
            net_result=[],
            port_result=[mock.Mock(id='PORT_ID', status='DOWN')],
            sg_result=[],
            floating_result=[],
            result={'port': 'PORT_ID'},
            exception=None,
            message='')),
        ('validate:net-y:port-n:fixed_ip-n:sgroups-y', dict(
            reason=None,
            success=True,
            inputs={'network': 'NET', 'security_groups': ['default']},
            net_result=[mock.Mock(id='NET_ID')],
            port_result=[],
            sg_result=[mock.Mock(id='SG_ID')],
            floating_result=[],
            result={'network': 'NET_ID', 'security_groups': ['SG_ID']},
            exception=None,
            message='')),
        ('validate:net-y:port-n:fixed_ip-n:sgroups-n:floating_net-y', dict(
            reason=None,
            success=True,
            inputs={'network': 'NET', 'floating_network': 'NET'},
            net_result=[mock.Mock(id='NET_ID'), mock.Mock(id='NET_ID')],
            port_result=[],
            sg_result=[],
            floating_result=[],
            result={'network': 'NET_ID', 'floating_network': 'NET_ID'},
            exception=None,
            message='')),
        ('validate:net-y:port-n:fixed_ip-n:floating_net-y:floating_ip-y', dict(
            reason=None,
            success=True,
            inputs={'network': 'NET', 'floating_network': 'NET',
                    'floating_ip': 'FLOATINGIP'},
            net_result=[mock.Mock(id='NET_ID'), mock.Mock(id='NET_ID')],
            port_result=[],
            sg_result=[],
            floating_result=[mock.Mock(id='FLOATINGIP_ID', status='INACTIVE')],
            result={'network': 'NET_ID', 'floating_network': 'NET_ID',
                    'floating_ip_id': 'FLOATINGIP_ID',
                    'floating_ip': 'FLOATINGIP'},
            exception=None,
            message='')),
        ('validate:net-y:port-n:fixed_ip-y:sgroups-n', dict(
            reason=None,
            success=True,
            inputs={'network': 'NET', 'fixed_ip': 'FIXED_IP'},
            net_result=[mock.Mock(id='NET_ID')],
            port_result=[],
            sg_result=[],
            floating_result=[],
            result={'network': 'NET_ID', 'fixed_ip': 'FIXED_IP'},
            exception=None,
            message='')),
        ('validate:net-f:port-y:fixed_ip-n:sgroups-n', dict(
            reason=None,
            success=False,
            inputs={'network': 'NET', 'port': 'PORT'},
            net_result=[exc.InternalError(message='NET Failure')],
            port_result=[],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.InvalidSpec,
            message='NET Failure')),
        ('validate:net-n:port-f:fixed_ip-n', dict(
            reason=None,
            success=False,
            inputs={'port': 'PORT'},
            net_result=[],
            port_result=[exc.InternalError(message='PORT Failure')],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.InvalidSpec,
            message='PORT Failure')),
        ('validate:net-n:port-active:fixed_ip-n', dict(
            reason=None,
            success=False,
            inputs={'port': 'PORT'},
            net_result=[],
            port_result=[mock.Mock(id='PORT_ID', status='ACTIVE')],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.InvalidSpec,
            message='The status of the port PORT must be DOWN')),
        ('validate:net-n:port-y:fixed_ip-n:floating_net-n:floating_ip-y', dict(
            reason=None,
            success=False,
            inputs={'port': 'PORT', 'floating_ip': 'FLOATINGIP'},
            net_result=[],
            port_result=[mock.Mock(id='PORT_ID', status='DOWN')],
            sg_result=[],
            floating_result=[mock.Mock(id='FLOATINGIP_ID', status='INACTIVE')],
            result={},
            exception=exc.InvalidSpec,
            message='Must specify a network to create floating IP')),
        ('validate:net-n:port-y:fixed_ip-n:floating_ip-active', dict(
            reason=None,
            success=False,
            inputs={'port': 'PORT', 'floating_network': 'NET',
                    'floating_ip': 'FLOATINGIP'},
            net_result=[mock.Mock(id='NET_ID')],
            port_result=[mock.Mock(id='PORT_ID', status='DOWN')],
            sg_result=[],
            floating_result=[mock.Mock(id='FLOATINGIP_ID', status='ACTIVE')],
            result={},
            exception=exc.InvalidSpec,
            message='the floating IP FLOATINGIP has been used.')),
        ('validate:net-n:port-n:fixed_ip-n', dict(
            reason=None,
            success=False,
            inputs={'fixed_ip': 'FIXED_IP'},
            net_result=[],
            port_result=[],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.InvalidSpec,
            message="One of 'port' and 'network' must be provided")),
        ('validate:net-n:port-y:fixed_ip-y', dict(
            reason=None,
            success=False,
            inputs={'port': 'PORT', 'fixed_ip': 'FIXED_IP'},
            net_result=[],
            port_result=[mock.Mock(id='PORT_ID', status='DOWN')],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.InvalidSpec,
            message=("The 'port' property and the 'fixed_ip' property cannot "
                     "be specified at the same time"))),
        ('create:net-y:port-y:fixed_ip-n', dict(
            reason='create',
            success=True,
            inputs={'network': 'NET', 'port': 'PORT'},
            net_result=[mock.Mock(id='NET_ID')],
            port_result=[mock.Mock(id='PORT_ID', status='DOWN')],
            sg_result=[],
            floating_result=[],
            result={'network': 'NET_ID', 'port': 'PORT_ID'},
            exception=None,
            message='')),
        ('create:net-y:port-n:fixed_ip-y', dict(
            reason='create',
            success=True,
            inputs={'network': 'NET', 'fixed_ip': 'FIXED_IP'},
            net_result=[mock.Mock(id='NET_ID')],
            port_result=[],
            sg_result=[],
            floating_result=[],
            result={'network': 'NET_ID', 'fixed_ip': 'FIXED_IP'},
            exception=None,
            message='')),
        ('create:net-y:port-n:fixed_ip-n:sgroups-y', dict(
            reason='create',
            success=True,
            inputs={'network': 'NET', 'security_groups': ['default']},
            net_result=[mock.Mock(id='NET_ID')],
            port_result=[],
            sg_result=[mock.Mock(id='SG_ID')],
            floating_result=[],
            result={'network': 'NET_ID', 'security_groups': ['SG_ID']},
            exception=None,
            message='')),
        ('create:net-y:port-n:fixed_ip-n:sgroups-n:floating_net-y', dict(
            reason=None,
            success=True,
            inputs={'network': 'NET', 'floating_network': 'NET'},
            net_result=[mock.Mock(id='NET_ID'), mock.Mock(id='NET_ID')],
            port_result=[],
            sg_result=[],
            floating_result=[],
            result={'network': 'NET_ID', 'floating_network': 'NET_ID'},
            exception=None,
            message='')),
        ('create:net-f:port-y:fixed_ip-n', dict(
            reason='create',
            success=False,
            inputs={'network': 'NET', 'port': 'PORT'},
            net_result=[exc.InternalError(message='NET Failure')],
            port_result=[],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.EResourceCreation,
            message='Failed in creating server: NET Failure.')),
        ('create:net-n:port-f:fixed_ip-n', dict(
            reason='create',
            success=False,
            inputs={'port': 'PORT'},
            net_result=[],
            port_result=[exc.InternalError(message='PORT Failure')],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.EResourceCreation,
            message='Failed in creating server: PORT Failure.')),
        ('create:net-n:port-active:fixed_ip-n', dict(
            reason='create',
            success=False,
            inputs={'port': 'PORT'},
            net_result=[],
            port_result=[mock.Mock(id='PORT_ID', status='ACTIVE')],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.EResourceCreation,
            message=('Failed in creating server: The status of the port PORT '
                     'must be DOWN.'))),
        ('create:net-n:port-n:fixed_ip-n', dict(
            reason='create',
            success=False,
            inputs={'fixed_ip': 'FIXED_IP'},
            net_result=[],
            port_result=[],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.EResourceCreation,
            message=("Failed in creating server: One of 'port' "
                     "and 'network' must be provided."))),
        ('create:net-n:port-y:fixed_ip-y', dict(
            reason='create',
            success=False,
            inputs={'port': 'PORT', 'fixed_ip': 'FIXED_IP'},
            net_result=[],
            port_result=[mock.Mock(id='PORT_ID', status='DOWN')],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.EResourceCreation,
            message=("Failed in creating server: The 'port' property and the "
                     "'fixed_ip' property cannot be specified at the same "
                     "time."))),
        ('update:net-y:port-y:fixed_ip-n', dict(
            reason='update',
            success=True,
            inputs={'network': 'NET', 'port': 'PORT'},
            net_result=[mock.Mock(id='NET_ID')],
            port_result=[mock.Mock(id='PORT_ID', status='DOWN')],
            sg_result=[],
            floating_result=[],
            result={'network': 'NET_ID', 'port': 'PORT_ID'},
            exception=None,
            message='')),
        ('update:net-y:port-n:fixed_ip-y', dict(
            reason='update',
            success=True,
            inputs={'network': 'NET', 'fixed_ip': 'FIXED_IP'},
            net_result=[mock.Mock(id='NET_ID')],
            port_result=[],
            sg_result=[],
            floating_result=[],
            result={'network': 'NET_ID',
                    'fixed_ip': 'FIXED_IP'},
            exception=None,
            message='')),
        ('update:net-y:port-n:fixed_ip-n:sgroups-y', dict(
            reason='create',
            success=True,
            inputs={'network': 'NET', 'security_groups': ['default']},
            net_result=[mock.Mock(id='NET_ID')],
            port_result=[],
            sg_result=[mock.Mock(id='SG_ID')],
            floating_result=[],
            result={'network': 'NET_ID', 'security_groups': ['SG_ID']},
            exception=None,
            message='')),
        ('update:net-y:port-n:fixed_ip-n:sgroups-n:floating_net-y', dict(
            reason=None,
            success=True,
            inputs={'network': 'NET', 'floating_network': 'NET'},
            net_result=[mock.Mock(id='NET_ID'), mock.Mock(id='NET_ID')],
            port_result=[],
            sg_result=[],
            floating_result=[],
            result={'network': 'NET_ID', 'floating_network': 'NET_ID'},
            exception=None,
            message='')),
        ('update:net-f:port-y:fixed_ip-n', dict(
            reason='update',
            success=False,
            inputs={'network': 'NET', 'port': 'PORT'},
            net_result=[exc.InternalError(message='NET Failure')],
            port_result=[],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.EResourceUpdate,
            message="Failed in updating server 'NOVA_ID': NET Failure.")),
        ('update:net-n:port-f:fixed_ip-n', dict(
            reason='update',
            success=False,
            inputs={'port': 'PORT'},
            net_result=[],
            port_result=[exc.InternalError(message='PORT Failure')],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.EResourceUpdate,
            message="Failed in updating server 'NOVA_ID': PORT Failure.")),
        ('update:net-n:port-active:fixed_ip-n', dict(
            reason='update',
            success=False,
            inputs={'port': 'PORT'},
            net_result=[],
            port_result=[mock.Mock(id='PORT_ID', status='ACTIVE')],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.EResourceUpdate,
            message=("Failed in updating server 'NOVA_ID': The status of the "
                     "port PORT must be DOWN."))),
        ('update:net-n:port-n:fixed_ip-n', dict(
            reason='update',
            success=False,
            inputs={'fixed_ip': 'FIXED_IP'},
            net_result=[],
            port_result=[],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.EResourceUpdate,
            message=("Failed in updating server 'NOVA_ID': One of 'port' "
                     "and 'network' must be provided."))),
        ('update:net-n:port-y:fixed_ip-y', dict(
            reason='update',
            success=False,
            inputs={'port': 'PORT', 'fixed_ip': 'FIXED_IP'},
            net_result=[],
            port_result=[mock.Mock(id='PORT_ID', status='DOWN')],
            sg_result=[],
            floating_result=[],
            result={},
            exception=exc.EResourceUpdate,
            message=("Failed in updating server 'NOVA_ID': The 'port' "
                     "property and the 'fixed_ip' property cannot be "
                     "specified at the same time."))),
    ]

    def setUp(self):
        super(TestNetworkValidation, self).setUp()

        self.nc = mock.Mock()
        self.profile = server.ServerProfile('t', spec)
        self.profile._networkclient = self.nc

    def test_validation(self):
        self.nc.network_get.side_effect = self.net_result
        self.nc.port_find.side_effect = self.port_result
        self.nc.security_group_find.side_effect = self.sg_result
        self.nc.floatingip_find.side_effect = self.floating_result
        obj = mock.Mock(physical_id='NOVA_ID')

        if self.success:
            res = self.profile._validate_network(obj, self.inputs, self.reason)
            self.assertEqual(self.result, res)
        else:
            ex = self.assertRaises(self.exception,
                                   self.profile._validate_network,
                                   obj, self.inputs, self.reason)
            self.assertEqual(self.message, six.text_type(ex))

        if self.net_result:
            self.nc.network_get.assert_called_with('NET')
        if self.port_result:
            self.nc.port_find.assert_called_once_with('PORT')
        if self.sg_result:
            self.nc.security_group_find.assert_called_once_with('default')
        if self.floating_result:
            self.nc.floatingip_find.assert_called_once_with('FLOATINGIP')


class TestNovaServerValidate(base.SenlinTestCase):

    def setUp(self):
        super(TestNovaServerValidate, self).setUp()

        self.context = utils.dummy_context()

    def test_do_validate_all_passed(self):
        profile = server.ServerProfile('t', spec)
        mock_az = self.patchobject(profile, '_validate_az')
        mock_flavor = self.patchobject(profile, '_validate_flavor')
        mock_image = self.patchobject(profile, '_validate_image')
        mock_keypair = self.patchobject(profile, '_validate_keypair')
        mock_network = self.patchobject(profile, '_validate_network')
        obj = mock.Mock()

        res = profile.do_validate(obj)

        properties = spec['properties']
        self.assertTrue(res)
        mock_az.assert_called_once_with(obj, properties['availability_zone'])
        mock_flavor.assert_called_once_with(obj, properties['flavor'])
        mock_image.assert_called_once_with(obj, properties['image'])
        mock_keypair.assert_called_once_with(obj, properties['key_name'])
        mock_network.assert_called_once_with(obj, properties['networks'][0])
